# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import datetime
import unittest

from aniso8601 import compat
from aniso8601.builders import (
    DatetimeTuple,
    DateTuple,
    DurationTuple,
    IntervalTuple,
    Limit,
    TimeTuple,
    TimezoneTuple,
)
from aniso8601.builders.python import (
    FractionalComponent,
    PythonTimeBuilder,
    _cast_to_fractional_component,
    fractional_range_check,
    year_range_check,
)
from aniso8601.exceptions import (
    DayOutOfBoundsError,
    HoursOutOfBoundsError,
    ISOFormatError,
    LeapSecondError,
    MidnightBoundsError,
    MinutesOutOfBoundsError,
    MonthOutOfBoundsError,
    SecondsOutOfBoundsError,
    WeekOutOfBoundsError,
    YearOutOfBoundsError,
)
from aniso8601.utcoffset import UTCOffset


class TestPythonTimeBuilder_UtiltyFunctions(unittest.TestCase):
    def test_year_range_check(self):
        yearlimit = Limit(
            "Invalid year string.",
            0000,
            9999,
            YearOutOfBoundsError,
            "Year must be between 1..9999.",
            None,
        )

        self.assertEqual(year_range_check("19", yearlimit), 1900)
        self.assertEqual(year_range_check("1234", yearlimit), 1234)
        self.assertEqual(year_range_check("1985", yearlimit), 1985)

    def test_fractional_range_check(self):
        limit = Limit(
            "Invalid string.", -1, 1, ValueError, "Value must be between -1..1.", None
        )

        self.assertEqual(fractional_range_check(10, "1", limit), 1)
        self.assertEqual(fractional_range_check(10, "-1", limit), -1)
        self.assertEqual(
            fractional_range_check(10, "0.1", limit), FractionalComponent(0, 1)
        )
        self.assertEqual(
            fractional_range_check(10, "-0.1", limit), FractionalComponent(-0, 1)
        )

        with self.assertRaises(ValueError):
            fractional_range_check(10, "1.1", limit)

        with self.assertRaises(ValueError):
            fractional_range_check(10, "-1.1", limit)

    def test_cast_to_fractional_component(self):
        self.assertEqual(
            _cast_to_fractional_component(10, "1.1"), FractionalComponent(1, 1)
        )
        self.assertEqual(
            _cast_to_fractional_component(10, "-1.1"), FractionalComponent(-1, 1)
        )

        self.assertEqual(
            _cast_to_fractional_component(100, "1.1"), FractionalComponent(1, 10)
        )
        self.assertEqual(
            _cast_to_fractional_component(100, "-1.1"), FractionalComponent(-1, 10)
        )


class TestPythonTimeBuilder(unittest.TestCase):
    def test_build_date(self):
        testtuples = (
            (
                {
                    "YYYY": "2013",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
                datetime.date(2013, 1, 1),
            ),
            (
                {
                    "YYYY": "0001",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
                datetime.date(1, 1, 1),
            ),
            (
                {
                    "YYYY": "19",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
                datetime.date(1900, 1, 1),
            ),
            (
                {
                    "YYYY": "10",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
                datetime.date(1000, 1, 1),
            ),
            (
                {
                    "YYYY": "1981",
                    "MM": "04",
                    "DD": "05",
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
                datetime.date(1981, 4, 5),
            ),
            (
                {
                    "YYYY": "1981",
                    "MM": "04",
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
                datetime.date(1981, 4, 1),
            ),
            (
                {
                    "YYYY": "1981",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": "095",
                },
                datetime.date(1981, 4, 5),
            ),
            (
                {
                    "YYYY": "1981",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": "365",
                },
                datetime.date(1981, 12, 31),
            ),
            (
                {
                    "YYYY": "1980",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": "366",
                },
                datetime.date(1980, 12, 31),
            ),
            # Make sure we shift in zeros
            (
                {
                    "YYYY": "1",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
                datetime.date(1000, 1, 1),
            ),
            (
                {
                    "YYYY": "12",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
                datetime.date(1200, 1, 1),
            ),
            (
                {
                    "YYYY": "123",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
                datetime.date(1230, 1, 1),
            ),
        )

        for testtuple in testtuples:
            result = PythonTimeBuilder.build_date(**testtuple[0])
            self.assertEqual(result, testtuple[1])

        # Test weekday
        testtuples = (
            (
                {
                    "YYYY": "2004",
                    "MM": None,
                    "DD": None,
                    "Www": "53",
                    "D": None,
                    "DDD": None,
                },
                datetime.date(2004, 12, 27),
                0,
            ),
            (
                {
                    "YYYY": "2009",
                    "MM": None,
                    "DD": None,
                    "Www": "01",
                    "D": None,
                    "DDD": None,
                },
                datetime.date(2008, 12, 29),
                0,
            ),
            (
                {
                    "YYYY": "2010",
                    "MM": None,
                    "DD": None,
                    "Www": "01",
                    "D": None,
                    "DDD": None,
                },
                datetime.date(2010, 1, 4),
                0,
            ),
            (
                {
                    "YYYY": "2009",
                    "MM": None,
                    "DD": None,
                    "Www": "53",
                    "D": None,
                    "DDD": None,
                },
                datetime.date(2009, 12, 28),
                0,
            ),
            (
                {
                    "YYYY": "2009",
                    "MM": None,
                    "DD": None,
                    "Www": "01",
                    "D": "1",
                    "DDD": None,
                },
                datetime.date(2008, 12, 29),
                0,
            ),
            (
                {
                    "YYYY": "2009",
                    "MM": None,
                    "DD": None,
                    "Www": "53",
                    "D": "7",
                    "DDD": None,
                },
                datetime.date(2010, 1, 3),
                6,
            ),
            (
                {
                    "YYYY": "2010",
                    "MM": None,
                    "DD": None,
                    "Www": "01",
                    "D": "1",
                    "DDD": None,
                },
                datetime.date(2010, 1, 4),
                0,
            ),
            (
                {
                    "YYYY": "2004",
                    "MM": None,
                    "DD": None,
                    "Www": "53",
                    "D": "6",
                    "DDD": None,
                },
                datetime.date(2005, 1, 1),
                5,
            ),
        )

        for testtuple in testtuples:
            result = PythonTimeBuilder.build_date(**testtuple[0])
            self.assertEqual(result, testtuple[1])
            self.assertEqual(result.weekday(), testtuple[2])

    def test_build_time(self):
        testtuples = (
            ({}, datetime.time()),
            ({"hh": "12.5"}, datetime.time(hour=12, minute=30)),
            (
                {"hh": "23.99999999997"},
                datetime.time(hour=23, minute=59, second=59, microsecond=999999),
            ),
            ({"hh": "1", "mm": "23"}, datetime.time(hour=1, minute=23)),
            (
                {"hh": "1", "mm": "23.4567"},
                datetime.time(hour=1, minute=23, second=27, microsecond=402000),
            ),
            (
                {"hh": "14", "mm": "43.999999997"},
                datetime.time(hour=14, minute=43, second=59, microsecond=999999),
            ),
            (
                {"hh": "1", "mm": "23", "ss": "45"},
                datetime.time(hour=1, minute=23, second=45),
            ),
            (
                {"hh": "23", "mm": "21", "ss": "28.512400"},
                datetime.time(hour=23, minute=21, second=28, microsecond=512400),
            ),
            (
                {"hh": "01", "mm": "03", "ss": "11.858714"},
                datetime.time(hour=1, minute=3, second=11, microsecond=858714),
            ),
            (
                {"hh": "14", "mm": "43", "ss": "59.9999997"},
                datetime.time(hour=14, minute=43, second=59, microsecond=999999),
            ),
            ({"hh": "24"}, datetime.time(hour=0)),
            ({"hh": "24", "mm": "00"}, datetime.time(hour=0)),
            ({"hh": "24", "mm": "00", "ss": "00"}, datetime.time(hour=0)),
            (
                {"tz": TimezoneTuple(False, None, "00", "00", "UTC")},
                datetime.time(tzinfo=UTCOffset(name="UTC", minutes=0)),
            ),
            (
                {
                    "hh": "23",
                    "mm": "21",
                    "ss": "28.512400",
                    "tz": TimezoneTuple(False, None, "00", "00", "+00:00"),
                },
                datetime.time(
                    hour=23,
                    minute=21,
                    second=28,
                    microsecond=512400,
                    tzinfo=UTCOffset(name="+00:00", minutes=0),
                ),
            ),
            (
                {
                    "hh": "1",
                    "mm": "23",
                    "tz": TimezoneTuple(False, None, "01", "00", "+1"),
                },
                datetime.time(
                    hour=1, minute=23, tzinfo=UTCOffset(name="+1", minutes=60)
                ),
            ),
            (
                {
                    "hh": "1",
                    "mm": "23.4567",
                    "tz": TimezoneTuple(True, None, "01", "00", "-1"),
                },
                datetime.time(
                    hour=1,
                    minute=23,
                    second=27,
                    microsecond=402000,
                    tzinfo=UTCOffset(name="-1", minutes=-60),
                ),
            ),
            (
                {
                    "hh": "23",
                    "mm": "21",
                    "ss": "28.512400",
                    "tz": TimezoneTuple(False, None, "01", "30", "+1:30"),
                },
                datetime.time(
                    hour=23,
                    minute=21,
                    second=28,
                    microsecond=512400,
                    tzinfo=UTCOffset(name="+1:30", minutes=90),
                ),
            ),
            (
                {
                    "hh": "23",
                    "mm": "21",
                    "ss": "28.512400",
                    "tz": TimezoneTuple(False, None, "11", "15", "+11:15"),
                },
                datetime.time(
                    hour=23,
                    minute=21,
                    second=28,
                    microsecond=512400,
                    tzinfo=UTCOffset(name="+11:15", minutes=675),
                ),
            ),
            (
                {
                    "hh": "23",
                    "mm": "21",
                    "ss": "28.512400",
                    "tz": TimezoneTuple(False, None, "12", "34", "+12:34"),
                },
                datetime.time(
                    hour=23,
                    minute=21,
                    second=28,
                    microsecond=512400,
                    tzinfo=UTCOffset(name="+12:34", minutes=754),
                ),
            ),
            (
                {
                    "hh": "23",
                    "mm": "21",
                    "ss": "28.512400",
                    "tz": TimezoneTuple(False, None, "00", "00", "UTC"),
                },
                datetime.time(
                    hour=23,
                    minute=21,
                    second=28,
                    microsecond=512400,
                    tzinfo=UTCOffset(name="UTC", minutes=0),
                ),
            ),
            # Make sure we truncate, not round
            # https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
            # https://bitbucket.org/nielsenb/aniso8601/issues/21/sub-microsecond-precision-is-lost-when
            (
                {"hh": "14.9999999999999999"},
                datetime.time(hour=14, minute=59, second=59, microsecond=999999),
            ),
            ({"mm": "0.00000000999"}, datetime.time()),
            ({"mm": "0.0000000999"}, datetime.time(microsecond=5)),
            ({"ss": "0.0000001"}, datetime.time()),
            ({"ss": "2.0000048"}, datetime.time(second=2, microsecond=4)),
        )

        for testtuple in testtuples:
            result = PythonTimeBuilder.build_time(**testtuple[0])
            self.assertEqual(result, testtuple[1])

    def test_build_datetime(self):
        testtuples = (
            (
                (
                    DateTuple("2019", "06", "05", None, None, None),
                    TimeTuple("01", "03", "11.858714", None),
                ),
                datetime.datetime(
                    2019, 6, 5, hour=1, minute=3, second=11, microsecond=858714
                ),
            ),
            (
                (
                    DateTuple("1234", "02", "03", None, None, None),
                    TimeTuple("23", "21", "28.512400", None),
                ),
                datetime.datetime(
                    1234, 2, 3, hour=23, minute=21, second=28, microsecond=512400
                ),
            ),
            (
                (
                    DateTuple("1981", "04", "05", None, None, None),
                    TimeTuple(
                        "23",
                        "21",
                        "28.512400",
                        TimezoneTuple(False, None, "11", "15", "+11:15"),
                    ),
                ),
                datetime.datetime(
                    1981,
                    4,
                    5,
                    hour=23,
                    minute=21,
                    second=28,
                    microsecond=512400,
                    tzinfo=UTCOffset(name="+11:15", minutes=675),
                ),
            ),
        )

        for testtuple in testtuples:
            result = PythonTimeBuilder.build_datetime(*testtuple[0])
            self.assertEqual(result, testtuple[1])

    def test_build_duration(self):
        testtuples = (
            (
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6",
                },
                datetime.timedelta(days=428, hours=4, minutes=54, seconds=6),
            ),
            (
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6.5",
                },
                datetime.timedelta(days=428, hours=4, minutes=54, seconds=6.5),
            ),
            ({"PnY": "1", "PnM": "2", "PnD": "3"}, datetime.timedelta(days=428)),
            ({"PnY": "1", "PnM": "2", "PnD": "3.5"}, datetime.timedelta(days=428.5)),
            (
                {"TnH": "4", "TnM": "54", "TnS": "6.5"},
                datetime.timedelta(hours=4, minutes=54, seconds=6.5),
            ),
            (
                {"TnH": "1", "TnM": "3", "TnS": "11.858714"},
                datetime.timedelta(hours=1, minutes=3, seconds=11, microseconds=858714),
            ),
            (
                {"TnH": "4", "TnM": "54", "TnS": "28.512400"},
                datetime.timedelta(
                    hours=4, minutes=54, seconds=28, microseconds=512400
                ),
            ),
            # Make sure we truncate, not round
            # https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
            # https://bitbucket.org/nielsenb/aniso8601/issues/21/sub-microsecond-precision-is-lost-when
            (
                {"PnY": "1999.9999999999999999"},
                datetime.timedelta(days=729999, seconds=86399, microseconds=999999),
            ),
            (
                {"PnM": "1.9999999999999999"},
                datetime.timedelta(
                    days=59, hours=23, minutes=59, seconds=59, microseconds=999999
                ),
            ),
            (
                {"PnW": "1.9999999999999999"},
                datetime.timedelta(
                    days=13, hours=23, minutes=59, seconds=59, microseconds=999999
                ),
            ),
            (
                {"PnD": "1.9999999999999999"},
                datetime.timedelta(
                    days=1, hours=23, minutes=59, seconds=59, microseconds=999999
                ),
            ),
            (
                {"TnH": "14.9999999999999999"},
                datetime.timedelta(
                    hours=14, minutes=59, seconds=59, microseconds=999999
                ),
            ),
            ({"TnM": "0.00000000999"}, datetime.timedelta(0)),
            ({"TnM": "0.0000000999"}, datetime.timedelta(microseconds=5)),
            ({"TnS": "0.0000001"}, datetime.timedelta(0)),
            ({"TnS": "2.0000048"}, datetime.timedelta(seconds=2, microseconds=4)),
            ({"PnY": "1"}, datetime.timedelta(days=365)),
            ({"PnY": "1.5"}, datetime.timedelta(days=547.5)),
            ({"PnM": "1"}, datetime.timedelta(days=30)),
            ({"PnM": "1.5"}, datetime.timedelta(days=45)),
            ({"PnW": "1"}, datetime.timedelta(days=7)),
            ({"PnW": "1.5"}, datetime.timedelta(days=10.5)),
            ({"PnD": "1"}, datetime.timedelta(days=1)),
            ({"PnD": "1.5"}, datetime.timedelta(days=1.5)),
            (
                {
                    "PnY": "0003",
                    "PnM": "06",
                    "PnD": "04",
                    "TnH": "12",
                    "TnM": "30",
                    "TnS": "05",
                },
                datetime.timedelta(days=1279, hours=12, minutes=30, seconds=5),
            ),
            (
                {
                    "PnY": "0003",
                    "PnM": "06",
                    "PnD": "04",
                    "TnH": "12",
                    "TnM": "30",
                    "TnS": "05.5",
                },
                datetime.timedelta(days=1279, hours=12, minutes=30, seconds=5.5),
            ),
            # Test timedelta limit
            (
                {"PnD": "999999999", "TnH": "23", "TnM": "59", "TnS": "59.999999"},
                datetime.timedelta.max,
            ),
            # Make sure we truncate, not round
            # https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
            (
                {
                    "PnY": "0001",
                    "PnM": "02",
                    "PnD": "03",
                    "TnH": "14",
                    "TnM": "43",
                    "TnS": "59.9999997",
                },
                datetime.timedelta(
                    days=428, hours=14, minutes=43, seconds=59, microseconds=999999
                ),
            ),
            # Verify overflows
            ({"TnH": "36"}, datetime.timedelta(days=1, hours=12)),
        )

        for testtuple in testtuples:
            result = PythonTimeBuilder.build_duration(**testtuple[0])
            self.assertEqual(result, testtuple[1])

    def test_build_interval(self):
        testtuples = (
            (
                {
                    "end": DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    "duration": DurationTuple(None, "1", None, None, None, None, None),
                },
                datetime.datetime(year=1981, month=4, day=5, hour=1, minute=1),
                datetime.datetime(year=1981, month=3, day=6, hour=1, minute=1),
            ),
            (
                {
                    "end": DateTuple("1981", "04", "05", None, None, None),
                    "duration": DurationTuple(None, "1", None, None, None, None, None),
                },
                datetime.date(year=1981, month=4, day=5),
                datetime.date(year=1981, month=3, day=6),
            ),
            (
                {
                    "end": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        "1.5", None, None, None, None, None, None
                    ),
                },
                datetime.date(year=2018, month=3, day=6),
                datetime.datetime(year=2016, month=9, day=4, hour=12),
            ),
            (
                {
                    "end": DateTuple("2014", "11", "12", None, None, None),
                    "duration": DurationTuple(None, None, None, None, "1", None, None),
                },
                datetime.date(year=2014, month=11, day=12),
                datetime.datetime(year=2014, month=11, day=11, hour=23),
            ),
            (
                {
                    "end": DateTuple("2014", "11", "12", None, None, None),
                    "duration": DurationTuple(None, None, None, None, "4", "54", "6.5"),
                },
                datetime.date(year=2014, month=11, day=12),
                datetime.datetime(
                    year=2014,
                    month=11,
                    day=11,
                    hour=19,
                    minute=5,
                    second=53,
                    microsecond=500000,
                ),
            ),
            (
                {
                    "end": DatetimeTuple(
                        DateTuple("2050", "03", "01", None, None, None),
                        TimeTuple(
                            "13",
                            "00",
                            "00",
                            TimezoneTuple(False, True, None, None, "Z"),
                        ),
                    ),
                    "duration": DurationTuple(None, None, None, None, "10", None, None),
                },
                datetime.datetime(
                    year=2050,
                    month=3,
                    day=1,
                    hour=13,
                    tzinfo=UTCOffset(name="UTC", minutes=0),
                ),
                datetime.datetime(
                    year=2050,
                    month=3,
                    day=1,
                    hour=3,
                    tzinfo=UTCOffset(name="UTC", minutes=0),
                ),
            ),
            # Make sure we truncate, not round
            # https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
            # https://bitbucket.org/nielsenb/aniso8601/issues/21/sub-microsecond-precision-is-lost-when
            (
                {
                    "end": DateTuple("2000", "01", "01", None, None, None),
                    "duration": DurationTuple(
                        "1999.9999999999999999", None, None, None, None, None, None
                    ),
                },
                datetime.date(year=2000, month=1, day=1),
                datetime.datetime(
                    year=1, month=4, day=30, hour=0, minute=0, second=0, microsecond=1
                ),
            ),
            (
                {
                    "end": DateTuple("1989", "03", "01", None, None, None),
                    "duration": DurationTuple(
                        None, "1.9999999999999999", None, None, None, None, None
                    ),
                },
                datetime.date(year=1989, month=3, day=1),
                datetime.datetime(
                    year=1988,
                    month=12,
                    day=31,
                    hour=0,
                    minute=0,
                    second=0,
                    microsecond=1,
                ),
            ),
            (
                {
                    "end": DateTuple("1989", "03", "01", None, None, None),
                    "duration": DurationTuple(
                        None, None, "1.9999999999999999", None, None, None, None
                    ),
                },
                datetime.date(year=1989, month=3, day=1),
                datetime.datetime(
                    year=1989,
                    month=2,
                    day=15,
                    hour=0,
                    minute=0,
                    second=0,
                    microsecond=1,
                ),
            ),
            (
                {
                    "end": DateTuple("1989", "03", "01", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, "1.9999999999999999", None, None, None
                    ),
                },
                datetime.date(year=1989, month=3, day=1),
                datetime.datetime(
                    year=1989,
                    month=2,
                    day=27,
                    hour=0,
                    minute=0,
                    second=0,
                    microsecond=1,
                ),
            ),
            (
                {
                    "end": DateTuple("2001", "01", "01", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, "14.9999999999999999", None, None
                    ),
                },
                datetime.date(year=2001, month=1, day=1),
                datetime.datetime(
                    year=2000,
                    month=12,
                    day=31,
                    hour=9,
                    minute=0,
                    second=0,
                    microsecond=1,
                ),
            ),
            (
                {
                    "end": DateTuple("2001", "01", "01", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, "0.00000000999", None
                    ),
                },
                datetime.date(year=2001, month=1, day=1),
                datetime.datetime(year=2001, month=1, day=1),
            ),
            (
                {
                    "end": DateTuple("2001", "01", "01", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, "0.0000000999", None
                    ),
                },
                datetime.date(year=2001, month=1, day=1),
                datetime.datetime(
                    year=2000,
                    month=12,
                    day=31,
                    hour=23,
                    minute=59,
                    second=59,
                    microsecond=999995,
                ),
            ),
            (
                {
                    "end": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, None, "0.0000001"
                    ),
                },
                datetime.date(year=2018, month=3, day=6),
                datetime.datetime(year=2018, month=3, day=6),
            ),
            (
                {
                    "end": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, None, "2.0000048"
                    ),
                },
                datetime.date(year=2018, month=3, day=6),
                datetime.datetime(
                    year=2018,
                    month=3,
                    day=5,
                    hour=23,
                    minute=59,
                    second=57,
                    microsecond=999996,
                ),
            ),
            (
                {
                    "start": DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    "duration": DurationTuple(None, "1", None, "1", None, "1", None),
                },
                datetime.datetime(year=1981, month=4, day=5, hour=1, minute=1),
                datetime.datetime(year=1981, month=5, day=6, hour=1, minute=2),
            ),
            (
                {
                    "start": DateTuple("1981", "04", "05", None, None, None),
                    "duration": DurationTuple(None, "1", None, "1", None, None, None),
                },
                datetime.date(year=1981, month=4, day=5),
                datetime.date(year=1981, month=5, day=6),
            ),
            (
                {
                    "start": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, "2.5", None, None, None, None, None
                    ),
                },
                datetime.date(year=2018, month=3, day=6),
                datetime.date(year=2018, month=5, day=20),
            ),
            (
                {
                    "start": DateTuple("2014", "11", "12", None, None, None),
                    "duration": DurationTuple(None, None, None, None, "1", None, None),
                },
                datetime.date(year=2014, month=11, day=12),
                datetime.datetime(year=2014, month=11, day=12, hour=1, minute=0),
            ),
            (
                {
                    "start": DateTuple("2014", "11", "12", None, None, None),
                    "duration": DurationTuple(None, None, None, None, "4", "54", "6.5"),
                },
                datetime.date(year=2014, month=11, day=12),
                datetime.datetime(
                    year=2014,
                    month=11,
                    day=12,
                    hour=4,
                    minute=54,
                    second=6,
                    microsecond=500000,
                ),
            ),
            (
                {
                    "start": DatetimeTuple(
                        DateTuple("2050", "03", "01", None, None, None),
                        TimeTuple(
                            "13",
                            "00",
                            "00",
                            TimezoneTuple(False, True, None, None, "Z"),
                        ),
                    ),
                    "duration": DurationTuple(None, None, None, None, "10", None, None),
                },
                datetime.datetime(
                    year=2050,
                    month=3,
                    day=1,
                    hour=13,
                    tzinfo=UTCOffset(name="UTC", minutes=0),
                ),
                datetime.datetime(
                    year=2050,
                    month=3,
                    day=1,
                    hour=23,
                    tzinfo=UTCOffset(name="UTC", minutes=0),
                ),
            ),
            # Make sure we truncate, not round
            # https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
            (
                {
                    "start": DateTuple("0001", "01", "01", None, None, None),
                    "duration": DurationTuple(
                        "1999.9999999999999999", None, None, None, None, None, None
                    ),
                },
                datetime.date(year=1, month=1, day=1),
                datetime.datetime(
                    year=1999,
                    month=9,
                    day=3,
                    hour=23,
                    minute=59,
                    second=59,
                    microsecond=999999,
                ),
            ),
            (
                {
                    "start": DateTuple("1989", "03", "01", None, None, None),
                    "duration": DurationTuple(
                        None, "1.9999999999999999", None, None, None, None, None
                    ),
                },
                datetime.date(year=1989, month=3, day=1),
                datetime.datetime(
                    year=1989,
                    month=4,
                    day=29,
                    hour=23,
                    minute=59,
                    second=59,
                    microsecond=999999,
                ),
            ),
            (
                {
                    "start": DateTuple("1989", "03", "01", None, None, None),
                    "duration": DurationTuple(
                        None, None, "1.9999999999999999", None, None, None, None
                    ),
                },
                datetime.date(year=1989, month=3, day=1),
                datetime.datetime(
                    year=1989,
                    month=3,
                    day=14,
                    hour=23,
                    minute=59,
                    second=59,
                    microsecond=999999,
                ),
            ),
            (
                {
                    "start": DateTuple("1989", "03", "01", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, "1.9999999999999999", None, None, None
                    ),
                },
                datetime.date(year=1989, month=3, day=1),
                datetime.datetime(
                    year=1989,
                    month=3,
                    day=2,
                    hour=23,
                    minute=59,
                    second=59,
                    microsecond=999999,
                ),
            ),
            (
                {
                    "start": DateTuple("2001", "01", "01", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, "14.9999999999999999", None, None
                    ),
                },
                datetime.date(year=2001, month=1, day=1),
                datetime.datetime(
                    year=2001,
                    month=1,
                    day=1,
                    hour=14,
                    minute=59,
                    second=59,
                    microsecond=999999,
                ),
            ),
            (
                {
                    "start": DateTuple("2001", "01", "01", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, "0.00000000999", None
                    ),
                },
                datetime.date(year=2001, month=1, day=1),
                datetime.datetime(year=2001, month=1, day=1),
            ),
            (
                {
                    "start": DateTuple("2001", "01", "01", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, "0.0000000999", None
                    ),
                },
                datetime.date(year=2001, month=1, day=1),
                datetime.datetime(
                    year=2001, month=1, day=1, hour=0, minute=0, second=0, microsecond=5
                ),
            ),
            (
                {
                    "start": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, None, "0.0000001"
                    ),
                },
                datetime.date(year=2018, month=3, day=6),
                datetime.datetime(year=2018, month=3, day=6),
            ),
            (
                {
                    "start": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, None, "2.0000048"
                    ),
                },
                datetime.date(year=2018, month=3, day=6),
                datetime.datetime(
                    year=2018, month=3, day=6, hour=0, minute=0, second=2, microsecond=4
                ),
            ),
            (
                {
                    "start": DatetimeTuple(
                        DateTuple("1980", "03", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    "end": DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                },
                datetime.datetime(year=1980, month=3, day=5, hour=1, minute=1),
                datetime.datetime(year=1981, month=4, day=5, hour=1, minute=1),
            ),
            (
                {
                    "start": DatetimeTuple(
                        DateTuple("1980", "03", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    "end": DateTuple("1981", "04", "05", None, None, None),
                },
                datetime.datetime(year=1980, month=3, day=5, hour=1, minute=1),
                datetime.date(year=1981, month=4, day=5),
            ),
            (
                {
                    "start": DateTuple("1980", "03", "05", None, None, None),
                    "end": DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                },
                datetime.date(year=1980, month=3, day=5),
                datetime.datetime(year=1981, month=4, day=5, hour=1, minute=1),
            ),
            (
                {
                    "start": DateTuple("1980", "03", "05", None, None, None),
                    "end": DateTuple("1981", "04", "05", None, None, None),
                },
                datetime.date(year=1980, month=3, day=5),
                datetime.date(year=1981, month=4, day=5),
            ),
            (
                {
                    "start": DateTuple("1981", "04", "05", None, None, None),
                    "end": DateTuple("1980", "03", "05", None, None, None),
                },
                datetime.date(year=1981, month=4, day=5),
                datetime.date(year=1980, month=3, day=5),
            ),
            (
                {
                    "start": DatetimeTuple(
                        DateTuple("2050", "03", "01", None, None, None),
                        TimeTuple(
                            "13",
                            "00",
                            "00",
                            TimezoneTuple(False, True, None, None, "Z"),
                        ),
                    ),
                    "end": DatetimeTuple(
                        DateTuple("2050", "05", "11", None, None, None),
                        TimeTuple(
                            "15",
                            "30",
                            "00",
                            TimezoneTuple(False, True, None, None, "Z"),
                        ),
                    ),
                },
                datetime.datetime(
                    year=2050,
                    month=3,
                    day=1,
                    hour=13,
                    tzinfo=UTCOffset(name="UTC", minutes=0),
                ),
                datetime.datetime(
                    year=2050,
                    month=5,
                    day=11,
                    hour=15,
                    minute=30,
                    tzinfo=UTCOffset(name="UTC", minutes=0),
                ),
            ),
            # Test concise representation
            (
                {
                    "start": DateTuple("2020", "01", "01", None, None, None),
                    "end": DateTuple(None, None, "02", None, None, None),
                },
                datetime.date(year=2020, month=1, day=1),
                datetime.date(year=2020, month=1, day=2),
            ),
            (
                {
                    "start": DateTuple("2008", "02", "15", None, None, None),
                    "end": DateTuple(None, "03", "14", None, None, None),
                },
                datetime.date(year=2008, month=2, day=15),
                datetime.date(year=2008, month=3, day=14),
            ),
            (
                {
                    "start": DatetimeTuple(
                        DateTuple("2007", "12", "14", None, None, None),
                        TimeTuple("13", "30", None, None),
                    ),
                    "end": TimeTuple("15", "30", None, None),
                },
                datetime.datetime(year=2007, month=12, day=14, hour=13, minute=30),
                datetime.datetime(year=2007, month=12, day=14, hour=15, minute=30),
            ),
            (
                {
                    "start": DatetimeTuple(
                        DateTuple("2007", "11", "13", None, None, None),
                        TimeTuple("09", "00", None, None),
                    ),
                    "end": DatetimeTuple(
                        DateTuple(None, None, "15", None, None, None),
                        TimeTuple("17", "00", None, None),
                    ),
                },
                datetime.datetime(year=2007, month=11, day=13, hour=9),
                datetime.datetime(year=2007, month=11, day=15, hour=17),
            ),
            (
                {
                    "start": DatetimeTuple(
                        DateTuple("2007", "11", "13", None, None, None),
                        TimeTuple("00", "00", None, None),
                    ),
                    "end": DatetimeTuple(
                        DateTuple(None, None, "16", None, None, None),
                        TimeTuple("00", "00", None, None),
                    ),
                },
                datetime.datetime(year=2007, month=11, day=13),
                datetime.datetime(year=2007, month=11, day=16),
            ),
            (
                {
                    "start": DatetimeTuple(
                        DateTuple("2007", "11", "13", None, None, None),
                        TimeTuple(
                            "09",
                            "00",
                            None,
                            TimezoneTuple(False, True, None, None, "Z"),
                        ),
                    ),
                    "end": DatetimeTuple(
                        DateTuple(None, None, "15", None, None, None),
                        TimeTuple("17", "00", None, None),
                    ),
                },
                datetime.datetime(
                    year=2007,
                    month=11,
                    day=13,
                    hour=9,
                    tzinfo=UTCOffset(name="UTC", minutes=0),
                ),
                datetime.datetime(
                    year=2007,
                    month=11,
                    day=15,
                    hour=17,
                    tzinfo=UTCOffset(name="UTC", minutes=0),
                ),
            ),
            (
                {
                    "start": DatetimeTuple(
                        DateTuple("2007", "11", "13", None, None, None),
                        TimeTuple("09", "00", None, None),
                    ),
                    "end": TimeTuple("12", "34.567", None, None),
                },
                datetime.datetime(year=2007, month=11, day=13, hour=9),
                datetime.datetime(
                    year=2007,
                    month=11,
                    day=13,
                    hour=12,
                    minute=34,
                    second=34,
                    microsecond=20000,
                ),
            ),
            (
                {
                    "start": DateTuple("2007", "11", "13", None, None, None),
                    "end": TimeTuple("12", "34", None, None),
                },
                datetime.date(year=2007, month=11, day=13),
                datetime.datetime(year=2007, month=11, day=13, hour=12, minute=34),
            ),
            # Make sure we truncate, not round
            # https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
            (
                {
                    "start": DatetimeTuple(
                        DateTuple("1980", "03", "05", None, None, None),
                        TimeTuple("01", "01", "00.0000001", None),
                    ),
                    "end": DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("14", "43", "59.9999997", None),
                    ),
                },
                datetime.datetime(year=1980, month=3, day=5, hour=1, minute=1),
                datetime.datetime(
                    year=1981,
                    month=4,
                    day=5,
                    hour=14,
                    minute=43,
                    second=59,
                    microsecond=999999,
                ),
            ),
        )

        for testtuple in testtuples:
            result = PythonTimeBuilder.build_interval(**testtuple[0])
            self.assertEqual(result[0], testtuple[1])
            self.assertEqual(result[1], testtuple[2])

    def test_build_repeating_interval(self):
        args = {
            "Rnn": "3",
            "interval": IntervalTuple(
                DateTuple("1981", "04", "05", None, None, None),
                None,
                DurationTuple(None, None, None, "1", None, None, None),
            ),
        }
        results = list(PythonTimeBuilder.build_repeating_interval(**args))

        self.assertEqual(results[0], datetime.date(year=1981, month=4, day=5))
        self.assertEqual(results[1], datetime.date(year=1981, month=4, day=6))
        self.assertEqual(results[2], datetime.date(year=1981, month=4, day=7))

        args = {
            "Rnn": "11",
            "interval": IntervalTuple(
                None,
                DatetimeTuple(
                    DateTuple("1980", "03", "05", None, None, None),
                    TimeTuple("01", "01", "00", None),
                ),
                DurationTuple(None, None, None, None, "1", "2", None),
            ),
        }
        results = list(PythonTimeBuilder.build_repeating_interval(**args))

        for dateindex in compat.range(0, 11):
            self.assertEqual(
                results[dateindex],
                datetime.datetime(year=1980, month=3, day=5, hour=1, minute=1)
                - dateindex * datetime.timedelta(hours=1, minutes=2),
            )

        args = {
            "Rnn": "2",
            "interval": IntervalTuple(
                DatetimeTuple(
                    DateTuple("1980", "03", "05", None, None, None),
                    TimeTuple("01", "01", "00", None),
                ),
                DatetimeTuple(
                    DateTuple("1981", "04", "05", None, None, None),
                    TimeTuple("01", "01", "00", None),
                ),
                None,
            ),
        }
        results = list(PythonTimeBuilder.build_repeating_interval(**args))

        self.assertEqual(
            results[0], datetime.datetime(year=1980, month=3, day=5, hour=1, minute=1)
        )
        self.assertEqual(
            results[1], datetime.datetime(year=1981, month=4, day=5, hour=1, minute=1)
        )

        args = {
            "Rnn": "2",
            "interval": IntervalTuple(
                DatetimeTuple(
                    DateTuple("1980", "03", "05", None, None, None),
                    TimeTuple("01", "01", "00", None),
                ),
                DatetimeTuple(
                    DateTuple("1981", "04", "05", None, None, None),
                    TimeTuple("01", "01", "00", None),
                ),
                None,
            ),
        }
        results = list(PythonTimeBuilder.build_repeating_interval(**args))

        self.assertEqual(
            results[0], datetime.datetime(year=1980, month=3, day=5, hour=1, minute=1)
        )
        self.assertEqual(
            results[1], datetime.datetime(year=1981, month=4, day=5, hour=1, minute=1)
        )

        args = {
            "R": True,
            "interval": IntervalTuple(
                None,
                DatetimeTuple(
                    DateTuple("1980", "03", "05", None, None, None),
                    TimeTuple("01", "01", "00", None),
                ),
                DurationTuple(None, None, None, None, "1", "2", None),
            ),
        }
        resultgenerator = PythonTimeBuilder.build_repeating_interval(**args)

        # Test the first 11 generated
        for dateindex in compat.range(0, 11):
            self.assertEqual(
                next(resultgenerator),
                datetime.datetime(year=1980, month=3, day=5, hour=1, minute=1)
                - dateindex * datetime.timedelta(hours=1, minutes=2),
            )

        args = {
            "R": True,
            "interval": IntervalTuple(
                DateTuple("1981", "04", "05", None, None, None),
                None,
                DurationTuple(None, None, None, "1", None, None, None),
            ),
        }
        resultgenerator = PythonTimeBuilder.build_repeating_interval(**args)

        # Test the first 11 generated
        for dateindex in compat.range(0, 11):
            self.assertEqual(
                next(resultgenerator),
                (
                    datetime.datetime(year=1981, month=4, day=5, hour=0, minute=0)
                    + dateindex * datetime.timedelta(days=1)
                ).date(),
            )

    def test_build_timezone(self):
        testtuples = (
            ({"Z": True, "name": "Z"}, datetime.timedelta(hours=0), "UTC"),
            (
                {"negative": False, "hh": "00", "mm": "00", "name": "+00:00"},
                datetime.timedelta(hours=0),
                "+00:00",
            ),
            (
                {"negative": False, "hh": "01", "mm": "00", "name": "+01:00"},
                datetime.timedelta(hours=1),
                "+01:00",
            ),
            (
                {"negative": True, "hh": "01", "mm": "00", "name": "-01:00"},
                -datetime.timedelta(hours=1),
                "-01:00",
            ),
            (
                {"negative": False, "hh": "00", "mm": "12", "name": "+00:12"},
                datetime.timedelta(minutes=12),
                "+00:12",
            ),
            (
                {"negative": False, "hh": "01", "mm": "23", "name": "+01:23"},
                datetime.timedelta(hours=1, minutes=23),
                "+01:23",
            ),
            (
                {"negative": True, "hh": "01", "mm": "23", "name": "-01:23"},
                -datetime.timedelta(hours=1, minutes=23),
                "-01:23",
            ),
            (
                {"negative": False, "hh": "00", "name": "+00"},
                datetime.timedelta(hours=0),
                "+00",
            ),
            (
                {"negative": False, "hh": "01", "name": "+01"},
                datetime.timedelta(hours=1),
                "+01",
            ),
            (
                {"negative": True, "hh": "01", "name": "-01"},
                -datetime.timedelta(hours=1),
                "-01",
            ),
            (
                {"negative": False, "hh": "12", "name": "+12"},
                datetime.timedelta(hours=12),
                "+12",
            ),
            (
                {"negative": True, "hh": "12", "name": "-12"},
                -datetime.timedelta(hours=12),
                "-12",
            ),
        )

        for testtuple in testtuples:
            result = PythonTimeBuilder.build_timezone(**testtuple[0])
            self.assertEqual(result.utcoffset(None), testtuple[1])
            self.assertEqual(result.tzname(None), testtuple[2])

    def test_range_check_date(self):
        # 0 isn't a valid year for a Python builder
        with self.assertRaises(YearOutOfBoundsError):
            PythonTimeBuilder.build_date(YYYY="0000")

        # Leap year
        # https://bitbucket.org/nielsenb/aniso8601/issues/14/parsing-ordinal-dates-should-only-allow
        with self.assertRaises(DayOutOfBoundsError):
            PythonTimeBuilder.build_date(YYYY="1981", DDD="366")

    def test_range_check_time(self):
        # Hour 24 can only represent midnight
        with self.assertRaises(MidnightBoundsError):
            PythonTimeBuilder.build_time(hh="24", mm="00", ss="01")

        with self.assertRaises(MidnightBoundsError):
            PythonTimeBuilder.build_time(hh="24", mm="00.1")

        with self.assertRaises(MidnightBoundsError):
            PythonTimeBuilder.build_time(hh="24", mm="01")

        with self.assertRaises(MidnightBoundsError):
            PythonTimeBuilder.build_time(hh="24.1")

    def test_range_check_duration(self):
        with self.assertRaises(YearOutOfBoundsError):
            PythonTimeBuilder.build_duration(
                PnY=str((datetime.timedelta.max.days // 365) + 1)
            )

        with self.assertRaises(MonthOutOfBoundsError):
            PythonTimeBuilder.build_duration(
                PnM=str((datetime.timedelta.max.days // 30) + 1)
            )

        with self.assertRaises(DayOutOfBoundsError):
            PythonTimeBuilder.build_duration(PnD=str(datetime.timedelta.max.days + 1))

        with self.assertRaises(WeekOutOfBoundsError):
            PythonTimeBuilder.build_duration(
                PnW=str((datetime.timedelta.max.days // 7) + 1)
            )

        with self.assertRaises(HoursOutOfBoundsError):
            PythonTimeBuilder.build_duration(
                TnH=str((datetime.timedelta.max.days * 24) + 24)
            )

        with self.assertRaises(MinutesOutOfBoundsError):
            PythonTimeBuilder.build_duration(
                TnM=str((datetime.timedelta.max.days * 24 * 60) + 24 * 60)
            )

        with self.assertRaises(SecondsOutOfBoundsError):
            PythonTimeBuilder.build_duration(
                TnS=str((datetime.timedelta.max.days * 24 * 60 * 60) + 24 * 60 * 60)
            )

        # Split max range across all parts
        maxpart = datetime.timedelta.max.days // 7

        with self.assertRaises(DayOutOfBoundsError):
            PythonTimeBuilder.build_duration(
                PnY=str((maxpart // 365) + 1),
                PnM=str((maxpart // 30) + 1),
                PnD=str((maxpart + 1)),
                PnW=str((maxpart // 7) + 1),
                TnH=str((maxpart * 24) + 1),
                TnM=str((maxpart * 24 * 60) + 1),
                TnS=str((maxpart * 24 * 60 * 60) + 1),
            )

    def test_range_check_interval(self):
        with self.assertRaises(YearOutOfBoundsError):
            PythonTimeBuilder.build_interval(
                start=DateTuple("0007", None, None, None, None, None),
                duration=DurationTuple(
                    None, None, None, str(datetime.timedelta.max.days), None, None, None
                ),
            )

        with self.assertRaises(YearOutOfBoundsError):
            PythonTimeBuilder.build_interval(
                start=DatetimeTuple(
                    DateTuple("0007", None, None, None, None, None),
                    TimeTuple("1", None, None, None),
                ),
                duration=DurationTuple(
                    str(datetime.timedelta.max.days // 365),
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                ),
            )

        with self.assertRaises(YearOutOfBoundsError):
            PythonTimeBuilder.build_interval(
                end=DateTuple("0001", None, None, None, None, None),
                duration=DurationTuple("3", None, None, None, None, None, None),
            )

        with self.assertRaises(YearOutOfBoundsError):
            PythonTimeBuilder.build_interval(
                end=DatetimeTuple(
                    DateTuple("0001", None, None, None, None, None),
                    TimeTuple("1", None, None, None),
                ),
                duration=DurationTuple("2", None, None, None, None, None, None),
            )

    def test_build_week_date(self):
        weekdate = PythonTimeBuilder._build_week_date(2009, 1)
        self.assertEqual(weekdate, datetime.date(year=2008, month=12, day=29))

        weekdate = PythonTimeBuilder._build_week_date(2009, 53, isoday=7)
        self.assertEqual(weekdate, datetime.date(year=2010, month=1, day=3))

    def test_build_ordinal_date(self):
        ordinaldate = PythonTimeBuilder._build_ordinal_date(1981, 95)
        self.assertEqual(ordinaldate, datetime.date(year=1981, month=4, day=5))

    def test_iso_year_start(self):
        yearstart = PythonTimeBuilder._iso_year_start(2004)
        self.assertEqual(yearstart, datetime.date(year=2003, month=12, day=29))

        yearstart = PythonTimeBuilder._iso_year_start(2010)
        self.assertEqual(yearstart, datetime.date(year=2010, month=1, day=4))

        yearstart = PythonTimeBuilder._iso_year_start(2009)
        self.assertEqual(yearstart, datetime.date(year=2008, month=12, day=29))

    def test_date_generator(self):
        startdate = datetime.date(year=2018, month=8, day=29)
        timedelta = datetime.timedelta(days=1)
        iterations = 10

        generator = PythonTimeBuilder._date_generator(startdate, timedelta, iterations)

        results = list(generator)

        for dateindex in compat.range(0, 10):
            self.assertEqual(
                results[dateindex],
                datetime.date(year=2018, month=8, day=29)
                + dateindex * datetime.timedelta(days=1),
            )

    def test_date_generator_unbounded(self):
        startdate = datetime.date(year=2018, month=8, day=29)
        timedelta = datetime.timedelta(days=5)

        generator = PythonTimeBuilder._date_generator_unbounded(startdate, timedelta)

        # Check the first 10 results
        for dateindex in compat.range(0, 10):
            self.assertEqual(
                next(generator),
                datetime.date(year=2018, month=8, day=29)
                + dateindex * datetime.timedelta(days=5),
            )

    def test_distribute_microseconds(self):
        self.assertEqual(PythonTimeBuilder._distribute_microseconds(1, (), ()), (1,))
        self.assertEqual(
            PythonTimeBuilder._distribute_microseconds(11, (0,), (10,)), (1, 1)
        )
        self.assertEqual(
            PythonTimeBuilder._distribute_microseconds(211, (0, 0), (100, 10)),
            (2, 1, 1),
        )

        self.assertEqual(PythonTimeBuilder._distribute_microseconds(1, (), ()), (1,))
        self.assertEqual(
            PythonTimeBuilder._distribute_microseconds(11, (5,), (10,)), (6, 1)
        )
        self.assertEqual(
            PythonTimeBuilder._distribute_microseconds(211, (10, 5), (100, 10)),
            (12, 6, 1),
        )



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import sys

PY2 = sys.version_info[0] == 2

if PY2:  # pragma: no cover
    range = xrange  # pylint: disable=undefined-variable
else:
    range = range


def is_string(tocheck):
    # pylint: disable=undefined-variable
    if PY2:  # pragma: no cover
        return isinstance(tocheck, str) or isinstance(tocheck, unicode)

    return isinstance(tocheck, str)



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from aniso8601.builders import TupleBuilder
from aniso8601.builders.python import PythonTimeBuilder
from aniso8601.compat import is_string
from aniso8601.exceptions import ISOFormatError
from aniso8601.resolution import DateResolution


def get_date_resolution(isodatestr):
    # Valid string formats are:
    #
    # Y[YYY]
    # YYYY-MM-DD
    # YYYYMMDD
    # YYYY-MM
    # YYYY-Www
    # YYYYWww
    # YYYY-Www-D
    # YYYYWwwD
    # YYYY-DDD
    # YYYYDDD
    isodatetuple = parse_date(isodatestr, builder=TupleBuilder)

    if isodatetuple.DDD is not None:
        # YYYY-DDD
        # YYYYDDD
        return DateResolution.Ordinal

    if isodatetuple.D is not None:
        # YYYY-Www-D
        # YYYYWwwD
        return DateResolution.Weekday

    if isodatetuple.Www is not None:
        # YYYY-Www
        # YYYYWww
        return DateResolution.Week

    if isodatetuple.DD is not None:
        # YYYY-MM-DD
        # YYYYMMDD
        return DateResolution.Day

    if isodatetuple.MM is not None:
        # YYYY-MM
        return DateResolution.Month

    # Y[YYY]
    return DateResolution.Year


def parse_date(isodatestr, builder=PythonTimeBuilder):
    # Given a string in any ISO 8601 date format, return a datetime.date
    # object that corresponds to the given date. Valid string formats are:
    #
    # YY
    # YYYY
    # YYYY-MM-DD
    # YYYYMMDD
    # YYYY-MM
    # YYYY-Www
    # YYYYWww
    # YYYY-Www-D
    # YYYYWwwD
    # YYYY-DDD
    # YYYYDDD
    if is_string(isodatestr) is False:
        raise ValueError("Date must be string.")

    if isodatestr.startswith("+") or isodatestr.startswith("-"):
        raise NotImplementedError(
            "ISO 8601 extended year representation not supported."
        )

    if len(isodatestr) == 0 or isodatestr.count("-") > 2:
        raise ISOFormatError('"{0}" is not a valid ISO 8601 date.'.format(isodatestr))

    yearstr = None
    monthstr = None
    daystr = None
    weekstr = None
    weekdaystr = None
    ordinaldaystr = None

    if len(isodatestr) in (2, 4):
        # YY
        # YYYY
        yearstr = isodatestr
    elif "W" in isodatestr:
        if len(isodatestr) == 10:
            # YYYY-Www-D
            yearstr = isodatestr[0:4]
            weekstr = isodatestr[6:8]
            weekdaystr = isodatestr[9]
        elif len(isodatestr) == 8:
            if "-" in isodatestr:
                # YYYY-Www
                yearstr = isodatestr[0:4]
                weekstr = isodatestr[6:]
            else:
                # YYYYWwwD
                yearstr = isodatestr[0:4]
                weekstr = isodatestr[5:7]
                weekdaystr = isodatestr[7]
        elif len(isodatestr) == 7:
            # YYYYWww
            yearstr = isodatestr[0:4]
            weekstr = isodatestr[5:]
    elif len(isodatestr) == 7:
        if "-" in isodatestr:
            # YYYY-MM
            yearstr = isodatestr[0:4]
            monthstr = isodatestr[5:]
        else:
            # YYYYDDD
            yearstr = isodatestr[0:4]
            ordinaldaystr = isodatestr[4:]
    elif len(isodatestr) == 8:
        if "-" in isodatestr:
            # YYYY-DDD
            yearstr = isodatestr[0:4]
            ordinaldaystr = isodatestr[5:]
        else:
            # YYYYMMDD
            yearstr = isodatestr[0:4]
            monthstr = isodatestr[4:6]
            daystr = isodatestr[6:]
    elif len(isodatestr) == 10:
        # YYYY-MM-DD
        yearstr = isodatestr[0:4]
        monthstr = isodatestr[5:7]
        daystr = isodatestr[8:]
    else:
        raise ISOFormatError('"{0}" is not a valid ISO 8601 date.'.format(isodatestr))

    hascomponent = False

    for componentstr in [yearstr, monthstr, daystr, weekstr, weekdaystr, ordinaldaystr]:
        if componentstr is not None:
            hascomponent = True

            if componentstr.isdigit() is False:
                raise ISOFormatError(
                    '"{0}" is not a valid ISO 8601 date.'.format(isodatestr)
                )

    if hascomponent is False:
        raise ISOFormatError('"{0}" is not a valid ISO 8601 date.'.format(isodatestr))

    return builder.build_date(
        YYYY=yearstr,
        MM=monthstr,
        DD=daystr,
        Www=weekstr,
        D=weekdaystr,
        DDD=ordinaldaystr,
    )



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import datetime


class UTCOffset(datetime.tzinfo):
    def __init__(self, name=None, minutes=None):
        # We build an offset in this manner since the
        # tzinfo class must have an init
        # "method that can be called with no arguments"
        self._name = name

        if minutes is not None:
            self._utcdelta = datetime.timedelta(minutes=minutes)
        else:
            self._utcdelta = None

    def __repr__(self):
        if self._utcdelta >= datetime.timedelta(hours=0):
            return "+{0} UTC".format(self._utcdelta)

        # From the docs:
        # String representations of timedelta objects are normalized
        # similarly to their internal representation. This leads to
        # somewhat unusual results for negative timedeltas.

        # Clean this up for printing purposes
        # Negative deltas start at -1 day
        correcteddays = abs(self._utcdelta.days + 1)

        # Negative deltas have a positive seconds
        deltaseconds = (24 * 60 * 60) - self._utcdelta.seconds

        # (24 hours / day) * (60 minutes / hour) * (60 seconds / hour)
        days, remainder = divmod(deltaseconds, 24 * 60 * 60)

        # (1 hour) * (60 minutes / hour) * (60 seconds / hour)
        hours, remainder = divmod(remainder, 1 * 60 * 60)

        # (1 minute) * (60 seconds / minute)
        minutes, seconds = divmod(remainder, 1 * 60)

        # Add any remaining days to the correcteddays count
        correcteddays += days

        if correcteddays == 0:
            return "-{0}:{1:02}:{2:02} UTC".format(hours, minutes, seconds)
        if correcteddays == 1:
            return "-1 day, {0}:{1:02}:{2:02} UTC".format(hours, minutes, seconds)

        return "-{0} days, {1}:{2:02}:{3:02} UTC".format(
            correcteddays, hours, minutes, seconds
        )

    def utcoffset(self, dt):
        return self._utcdelta

    def tzname(self, dt):
        return self._name

    def dst(self, dt):
        # ISO 8601 specifies offsets should be different if DST is required,
        # instead of allowing for a DST to be specified
        # https://docs.python.org/2/library/datetime.html#datetime.tzinfo.dst
        return datetime.timedelta(0)



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import unittest

import aniso8601
from aniso8601.duration import (
    _has_any_component,
    _parse_duration_combined,
    _parse_duration_prescribed,
    _parse_duration_prescribed_notime,
    _parse_duration_prescribed_time,
    get_duration_resolution,
    parse_duration,
)
from aniso8601.exceptions import ISOFormatError
from aniso8601.resolution import DurationResolution
from aniso8601.tests.compat import mock


class TestDurationParserFunctions(unittest.TestCase):
    def test_get_duration_resolution_years(self):
        self.assertEqual(get_duration_resolution("P1Y"), DurationResolution.Years)
        self.assertEqual(get_duration_resolution("P1,5Y"), DurationResolution.Years)
        self.assertEqual(get_duration_resolution("P1.5Y"), DurationResolution.Years)

    def test_get_duration_resolution_months(self):
        self.assertEqual(get_duration_resolution("P1Y2M"), DurationResolution.Months)
        self.assertEqual(get_duration_resolution("P1M"), DurationResolution.Months)
        self.assertEqual(get_duration_resolution("P1,5M"), DurationResolution.Months)
        self.assertEqual(get_duration_resolution("P1.5M"), DurationResolution.Months)

    def test_get_duration_resolution_weeks(self):
        self.assertEqual(get_duration_resolution("P1W"), DurationResolution.Weeks)
        self.assertEqual(get_duration_resolution("P1,5W"), DurationResolution.Weeks)
        self.assertEqual(get_duration_resolution("P1.5W"), DurationResolution.Weeks)

    def test_get_duration_resolution_days(self):
        self.assertEqual(get_duration_resolution("P1Y2M3D"), DurationResolution.Days)
        self.assertEqual(get_duration_resolution("P1Y2M3,5D"), DurationResolution.Days)
        self.assertEqual(get_duration_resolution("P1Y2M3.5D"), DurationResolution.Days)
        self.assertEqual(get_duration_resolution("P1D"), DurationResolution.Days)
        self.assertEqual(get_duration_resolution("P1,5D"), DurationResolution.Days)
        self.assertEqual(get_duration_resolution("P1.5D"), DurationResolution.Days)

    def test_get_duration_resolution_hours(self):
        self.assertEqual(
            get_duration_resolution("P1Y2M3DT4H"), DurationResolution.Hours
        )
        self.assertEqual(get_duration_resolution("PT4H"), DurationResolution.Hours)

    def test_get_duration_resolution_minutes(self):
        self.assertEqual(
            get_duration_resolution("P1Y2M3DT4H5M"), DurationResolution.Minutes
        )
        self.assertEqual(get_duration_resolution("PT4H5M"), DurationResolution.Minutes)

    def test_get_duration_resolution_seconds(self):
        self.assertEqual(
            get_duration_resolution("P1Y2M3DT4H54M6S"), DurationResolution.Seconds
        )
        self.assertEqual(
            get_duration_resolution("P1Y2M3DT4H54M6,5S"), DurationResolution.Seconds
        )
        self.assertEqual(
            get_duration_resolution("P1Y2M3DT4H54M6.5S"), DurationResolution.Seconds
        )
        self.assertEqual(
            get_duration_resolution("PT4H54M6,5S"), DurationResolution.Seconds
        )
        self.assertEqual(
            get_duration_resolution("PT4H54M6.5S"), DurationResolution.Seconds
        )
        self.assertEqual(
            get_duration_resolution("PT0.0000001S"), DurationResolution.Seconds
        )
        self.assertEqual(
            get_duration_resolution("PT2.0000048S"), DurationResolution.Seconds
        )
        self.assertEqual(
            get_duration_resolution("P0003-06-04T12:30:05"), DurationResolution.Seconds
        )
        self.assertEqual(
            get_duration_resolution("P0003-06-04T12:30:05.5"),
            DurationResolution.Seconds,
        )
        self.assertEqual(
            get_duration_resolution("P0001-02-03T14:43:59.9999997"),
            DurationResolution.Seconds,
        )

    def test_parse_duration(self):
        testtuples = (
            (
                "P1Y2M3DT4H54M6S",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6",
                },
            ),
            (
                "P1Y2M3DT4H54M6,5S",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6.5",
                },
            ),
            (
                "P1Y2M3DT4H54M6.5S",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6.5",
                },
            ),
            (
                "P1YT4H",
                {
                    "PnY": "1",
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": None,
                    "TnS": None,
                },
            ),
            (
                "P1YT54M",
                {
                    "PnY": "1",
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": "54",
                    "TnS": None,
                },
            ),
            (
                "P1YT6S",
                {
                    "PnY": "1",
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": None,
                    "TnS": "6",
                },
            ),
            (
                "P1YT4H54M",
                {
                    "PnY": "1",
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": None,
                },
            ),
            (
                "P1YT4H6S",
                {
                    "PnY": "1",
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": None,
                    "TnS": "6",
                },
            ),
            (
                "P1YT54M6S",
                {
                    "PnY": "1",
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": "54",
                    "TnS": "6",
                },
            ),
            (
                "P1YT4H54M6S",
                {
                    "PnY": "1",
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6",
                },
            ),
            (
                "P2MT4H",
                {
                    "PnY": None,
                    "PnM": "2",
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": None,
                    "TnS": None,
                },
            ),
            (
                "P2MT54M",
                {
                    "PnY": None,
                    "PnM": "2",
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": "54",
                    "TnS": None,
                },
            ),
            (
                "P2MT6S",
                {
                    "PnY": None,
                    "PnM": "2",
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": None,
                    "TnS": "6",
                },
            ),
            (
                "P2MT4H54M",
                {
                    "PnY": None,
                    "PnM": "2",
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": None,
                },
            ),
            (
                "P2MT4H6S",
                {
                    "PnY": None,
                    "PnM": "2",
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": None,
                    "TnS": "6",
                },
            ),
            (
                "P2MT54M6S",
                {
                    "PnY": None,
                    "PnM": "2",
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": "54",
                    "TnS": "6",
                },
            ),
            (
                "P2MT4H54M6S",
                {
                    "PnY": None,
                    "PnM": "2",
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6",
                },
            ),
            (
                "P3DT4H",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": None,
                    "TnS": None,
                },
            ),
            (
                "P3DT54M",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": "3",
                    "TnH": None,
                    "TnM": "54",
                    "TnS": None,
                },
            ),
            (
                "P3DT6S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": "3",
                    "TnH": None,
                    "TnM": None,
                    "TnS": "6",
                },
            ),
            (
                "P3DT4H54M",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": None,
                },
            ),
            (
                "P3DT4H6S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": None,
                    "TnS": "6",
                },
            ),
            (
                "P3DT54M6S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": "3",
                    "TnH": None,
                    "TnM": "54",
                    "TnS": "6",
                },
            ),
            (
                "P3DT4H54M6S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6",
                },
            ),
            (
                "P1Y2MT4H",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": None,
                    "TnS": None,
                },
            ),
            (
                "P1Y2MT54M",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": "54",
                    "TnS": None,
                },
            ),
            (
                "P1Y2MT6S",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": None,
                    "TnS": "6",
                },
            ),
            (
                "P1Y2MT4H54M",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": None,
                },
            ),
            (
                "P1Y2MT4H6S",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": None,
                    "TnS": "6",
                },
            ),
            (
                "P1Y2MT54M6S",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": "54",
                    "TnS": "6",
                },
            ),
            (
                "P1Y2MT4H54M6S",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6",
                },
            ),
            (
                "P1Y3DT4H",
                {
                    "PnY": "1",
                    "PnM": None,
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": None,
                    "TnS": None,
                },
            ),
            (
                "P1Y3DT54M",
                {
                    "PnY": "1",
                    "PnM": None,
                    "PnW": None,
                    "PnD": "3",
                    "TnH": None,
                    "TnM": "54",
                    "TnS": None,
                },
            ),
            (
                "P1Y3DT6S",
                {
                    "PnY": "1",
                    "PnM": None,
                    "PnW": None,
                    "PnD": "3",
                    "TnH": None,
                    "TnM": None,
                    "TnS": "6",
                },
            ),
            (
                "P1Y3DT4H54M",
                {
                    "PnY": "1",
                    "PnM": None,
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": None,
                },
            ),
            (
                "P1Y3DT4H6S",
                {
                    "PnY": "1",
                    "PnM": None,
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": None,
                    "TnS": "6",
                },
            ),
            (
                "P1Y3DT54M6S",
                {
                    "PnY": "1",
                    "PnM": None,
                    "PnW": None,
                    "PnD": "3",
                    "TnH": None,
                    "TnM": "54",
                    "TnS": "6",
                },
            ),
            (
                "P1Y3DT4H54M6S",
                {
                    "PnY": "1",
                    "PnM": None,
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6",
                },
            ),
            (
                "P2M3DT4H",
                {
                    "PnY": None,
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": None,
                    "TnS": None,
                },
            ),
            (
                "P2M3DT54M",
                {
                    "PnY": None,
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": None,
                    "TnM": "54",
                    "TnS": None,
                },
            ),
            (
                "P2M3DT6S",
                {
                    "PnY": None,
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": None,
                    "TnM": None,
                    "TnS": "6",
                },
            ),
            (
                "P2M3DT4H54M",
                {
                    "PnY": None,
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": None,
                },
            ),
            (
                "P2M3DT4H6S",
                {
                    "PnY": None,
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": None,
                    "TnS": "6",
                },
            ),
            (
                "P2M3DT54M6S",
                {
                    "PnY": None,
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": None,
                    "TnM": "54",
                    "TnS": "6",
                },
            ),
            (
                "P2M3DT4H54M6S",
                {
                    "PnY": None,
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6",
                },
            ),
            (
                "PT4H54M6,5S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6.5",
                },
            ),
            (
                "PT4H54M6.5S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6.5",
                },
            ),
            (
                "PT4H",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": None,
                    "TnS": None,
                },
            ),
            (
                "PT5M",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": "5",
                    "TnS": None,
                },
            ),
            (
                "PT6S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": None,
                    "TnS": "6",
                },
            ),
            (
                "PT1H2M",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "1",
                    "TnM": "2",
                    "TnS": None,
                },
            ),
            (
                "PT3H4S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "3",
                    "TnM": None,
                    "TnS": "4",
                },
            ),
            (
                "PT5M6S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": "5",
                    "TnS": "6",
                },
            ),
            (
                "PT0.0000001S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": None,
                    "TnS": "0.0000001",
                },
            ),
            (
                "PT2.0000048S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": None,
                    "TnS": "2.0000048",
                },
            ),
            ("P1Y", {"PnY": "1", "PnM": None, "PnW": None, "PnD": None}),
            ("P1,5Y", {"PnY": "1.5", "PnM": None, "PnW": None, "PnD": None}),
            ("P1.5Y", {"PnY": "1.5", "PnM": None, "PnW": None, "PnD": None}),
            ("P1M", {"PnY": None, "PnM": "1", "PnW": None, "PnD": None}),
            ("P1,5M", {"PnY": None, "PnM": "1.5", "PnW": None, "PnD": None}),
            ("P1.5M", {"PnY": None, "PnM": "1.5", "PnW": None, "PnD": None}),
            ("P1W", {"PnY": None, "PnM": None, "PnW": "1", "PnD": None}),
            ("P1,5W", {"PnY": None, "PnM": None, "PnW": "1.5", "PnD": None}),
            ("P1.5W", {"PnY": None, "PnM": None, "PnW": "1.5", "PnD": None}),
            ("P1D", {"PnY": None, "PnM": None, "PnW": None, "PnD": "1"}),
            ("P1,5D", {"PnY": None, "PnM": None, "PnW": None, "PnD": "1.5"}),
            ("P1.5D", {"PnY": None, "PnM": None, "PnW": None, "PnD": "1.5"}),
            ("P1Y2M3D", {"PnY": "1", "PnM": "2", "PnW": None, "PnD": "3"}),
            ("P1Y2M3,5D", {"PnY": "1", "PnM": "2", "PnW": None, "PnD": "3.5"}),
            ("P1Y2M3.5D", {"PnY": "1", "PnM": "2", "PnW": None, "PnD": "3.5"}),
            ("P1Y2M", {"PnY": "1", "PnM": "2", "PnW": None, "PnD": None}),
            (
                "P0003-06-04T12:30:05",
                {
                    "PnY": "0003",
                    "PnM": "06",
                    "PnD": "04",
                    "TnH": "12",
                    "TnM": "30",
                    "TnS": "05",
                },
            ),
            (
                "P0003-06-04T12:30:05.5",
                {
                    "PnY": "0003",
                    "PnM": "06",
                    "PnD": "04",
                    "TnH": "12",
                    "TnM": "30",
                    "TnS": "05.5",
                },
            ),
            (
                "P0001-02-03T14:43:59.9999997",
                {
                    "PnY": "0001",
                    "PnM": "02",
                    "PnD": "03",
                    "TnH": "14",
                    "TnM": "43",
                    "TnS": "59.9999997",
                },
            ),
        )

        for testtuple in testtuples:
            with mock.patch.object(
                aniso8601.duration.PythonTimeBuilder, "build_duration"
            ) as mockBuildDuration:
                mockBuildDuration.return_value = testtuple[1]

                result = parse_duration(testtuple[0])

                self.assertEqual(result, testtuple[1])
                mockBuildDuration.assert_called_once_with(**testtuple[1])

    def test_parse_duration_mockbuilder(self):
        mockBuilder = mock.Mock()

        expectedargs = {
            "PnY": "1",
            "PnM": "2",
            "PnW": None,
            "PnD": "3",
            "TnH": "4",
            "TnM": "54",
            "TnS": "6",
        }

        mockBuilder.build_duration.return_value = expectedargs

        result = parse_duration("P1Y2M3DT4H54M6S", builder=mockBuilder)

        self.assertEqual(result, expectedargs)
        mockBuilder.build_duration.assert_called_once_with(**expectedargs)

    def test_parse_duration_badtype(self):
        testtuples = (None, 1, False, 1.234)

        for testtuple in testtuples:
            with self.assertRaises(ValueError):
                parse_duration(testtuple, builder=None)

    def test_parse_duration_nop(self):
        with self.assertRaises(ISOFormatError):
            # Duration must start with a P
            parse_duration("1Y2M3DT4H54M6S", builder=None)

    def test_parse_duration_weekcombination(self):
        # Week designator cannot be combined with other time designators
        # https://bitbucket.org/nielsenb/aniso8601/issues/2/week-designators-should-not-be-combinable

        with self.assertRaises(ISOFormatError):
            parse_duration("P1Y2W", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P1M2W", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P2W3D", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P1Y2W3D", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P1M2W3D", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P1Y1M2W3D", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P7WT4H", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P7WT54M", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P7WT6S", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P7WT4H54M", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P7WT4H6S", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P7WT54M6S", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P7WT4H54M6S", builder=None)

    def test_parse_duration_negative(self):
        with self.assertRaises(ISOFormatError):
            parse_duration("P-1Y", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P-2M", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P-3D", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P-T4H", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P-T54M", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P-T6S", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P-7W", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P-1Y2M3DT4H54M6S", builder=None)

    def test_parse_duration_outoforder(self):
        # Ensure durations are required to be in the correct order
        # https://bitbucket.org/nielsenb/aniso8601/issues/7/durations-with-time-components-before-t
        # https://bitbucket.org/nielsenb/aniso8601/issues/8/durations-with-components-in-wrong-order
        with self.assertRaises(ISOFormatError):
            parse_duration("P1S", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P1D1S", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P1H1M", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("1Y2M3D1SPT1M", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P1Y2M3D2MT1S", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P2M3D1ST1Y1M", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P1Y2M2MT3D1S", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("P1D1Y1M", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_duration("PT1S1H", builder=None)

    def test_parse_duration_badstr(self):
        testtuples = (
            "PPPPPPPPPPPPPPPPPPPPPPPPPPPP",
            "PTT",
            "PX7DDDTX8888UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU"
            "UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU8888888888888888H$H",
            "P1Y2M3X.4D",
            "P1Y2M3.4XD",
            "P1Y2M3DT4H5M6XS",
            "PT4H5M6X.2S",
            "bad",
            "",
        )

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                parse_duration(testtuple, builder=None)

    def test_parse_duration_prescribed(self):
        testtuples = (
            (
                "P1Y2M3DT4H54M6S",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6",
                },
            ),
            (
                "P1Y2M3DT4H54M6,5S",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6.5",
                },
            ),
            (
                "P1Y2M3DT4H54M6.5S",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6.5",
                },
            ),
            (
                "PT4H54M6,5S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6.5",
                },
            ),
            (
                "PT4H54M6.5S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6.5",
                },
            ),
            ("P1Y2M3D", {"PnY": "1", "PnM": "2", "" "PnW": None, "PnD": "3"}),
            ("P1Y2M3,5D", {"PnY": "1", "PnM": "2", "PnW": None, "PnD": "3.5"}),
            ("P1Y2M3.5D", {"PnY": "1", "PnM": "2", "PnW": None, "PnD": "3.5"}),
            ("P1Y2M", {"PnY": "1", "PnM": "2", "PnW": None, "PnD": None}),
            ("P1Y", {"PnY": "1", "PnM": None, "PnW": None, "PnD": None}),
            ("P1,5Y", {"PnY": "1.5", "PnM": None, "PnW": None, "PnD": None}),
            ("P1.5Y", {"PnY": "1.5", "PnM": None, "PnW": None, "PnD": None}),
            ("P1M", {"PnY": None, "PnM": "1", "PnW": None, "PnD": None}),
            ("P1,5M", {"PnY": None, "PnM": "1.5", "PnW": None, "PnD": None}),
            ("P1.5M", {"PnY": None, "PnM": "1.5", "PnW": None, "PnD": None}),
            ("P1W", {"PnY": None, "PnM": None, "PnW": "1", "PnD": None}),
            ("P1,5W", {"PnY": None, "PnM": None, "PnW": "1.5", "PnD": None}),
            ("P1.5W", {"PnY": None, "PnM": None, "PnW": "1.5", "PnD": None}),
            ("P1D", {"PnY": None, "PnM": None, "PnW": None, "PnD": "1"}),
            ("P1,5D", {"PnY": None, "PnM": None, "PnW": None, "PnD": "1.5"}),
            ("P1.5D", {"PnY": None, "PnM": None, "PnW": None, "PnD": "1.5"}),
        )

        for testtuple in testtuples:
            result = _parse_duration_prescribed(testtuple[0])

            self.assertEqual(result, testtuple[1])

    def test_parse_duration_prescribed_negative(self):
        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed("P-1Y")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed("P-2M")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed("P-3D")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed("P-4W")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed("P-1Y2M3D")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed("P-T1H")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed("P-T2M")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed("P-T3S")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed("P-1Y2M3DT4H54M6S")

    def test_parse_duration_prescribed_multiplefractions(self):
        with self.assertRaises(ISOFormatError):
            # Multiple fractions are not allowed
            _parse_duration_prescribed("P1Y2M3DT4H5.1234M6.1234S")

    def test_parse_duration_prescribed_middlefraction(self):
        with self.assertRaises(ISOFormatError):
            # Fraction only allowed on final component
            _parse_duration_prescribed("P1Y2M3DT4H5.1234M6S")

    def test_parse_duration_prescribed_suffixgarbage(self):
        # Don't allow garbage after the duration
        # https://bitbucket.org/nielsenb/aniso8601/issues/9/durations-with-trailing-garbage-are-parsed
        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed("P1Dasdfasdf")

    def test_parse_duration_prescribed_notime(self):
        testtuples = (
            ("P1Y2M3D", {"PnY": "1", "PnM": "2", "PnW": None, "PnD": "3"}),
            ("P1Y2M3,5D", {"PnY": "1", "PnM": "2", "PnW": None, "PnD": "3.5"}),
            ("P1Y2M3.5D", {"PnY": "1", "PnM": "2", "PnW": None, "PnD": "3.5"}),
            ("P1Y3D", {"PnY": "1", "PnM": None, "PnW": None, "PnD": "3"}),
            ("P1Y2M", {"PnY": "1", "PnM": "2", "PnW": None, "PnD": None}),
            ("P2M3D", {"PnY": None, "PnM": "2", "PnW": None, "PnD": "3"}),
            ("P1Y", {"PnY": "1", "PnM": None, "PnW": None, "PnD": None}),
            ("P1,5Y", {"PnY": "1.5", "PnM": None, "PnW": None, "PnD": None}),
            ("P1.5Y", {"PnY": "1.5", "PnM": None, "PnW": None, "PnD": None}),
            ("P1M", {"PnY": None, "PnM": "1", "PnW": None, "PnD": None}),
            ("P1,5M", {"PnY": None, "PnM": "1.5", "PnW": None, "PnD": None}),
            ("P1.5M", {"PnY": None, "PnM": "1.5", "PnW": None, "PnD": None}),
            ("P1W", {"PnY": None, "PnM": None, "PnW": "1", "PnD": None}),
            ("P1,5W", {"PnY": None, "PnM": None, "PnW": "1.5", "PnD": None}),
            ("P1.5W", {"PnY": None, "PnM": None, "PnW": "1.5", "PnD": None}),
            ("P1D", {"PnY": None, "PnM": None, "PnW": None, "PnD": "1"}),
            ("P1,5D", {"PnY": None, "PnM": None, "PnW": None, "PnD": "1.5"}),
            ("P1.5D", {"PnY": None, "PnM": None, "PnW": None, "PnD": "1.5"}),
        )

        for testtuple in testtuples:
            result = _parse_duration_prescribed_notime(testtuple[0])

            self.assertEqual(result, testtuple[1])

    def test_parse_duration_prescribed_notime_timepart(self):
        # Ensure no time part is allowed
        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_notime("P1S")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_notime("P1D1S")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_notime("P1H1M")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_notime("P1Y2M3D4H")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_notime("P1Y2M3D4H5S")

    def test_parse_duration_prescribed_notime_outoforder(self):
        # Ensure durations are required to be in the correct order
        # https://bitbucket.org/nielsenb/aniso8601/issues/8/durations-with-components-in-wrong-order
        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_notime("P1H1M")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_notime("P1D1Y1M")

    def test_parse_duration_prescribed_notime_badstr(self):
        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_notime("P1S")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_notime("P1D1S")

    def test_parse_duration_prescribed_time(self):
        testtuples = (
            (
                "P1Y2M3DT4H54M6S",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6",
                },
            ),
            (
                "P1Y2M3DT4H54M6,5S",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6.5",
                },
            ),
            (
                "P1Y2M3DT4H54M6.5S",
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": None,
                    "PnD": "3",
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6.5",
                },
            ),
            (
                "PT4H54M6,5S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6.5",
                },
            ),
            (
                "PT4H54M6.5S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": "54",
                    "TnS": "6.5",
                },
            ),
            (
                "PT4H",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "4",
                    "TnM": None,
                    "TnS": None,
                },
            ),
            (
                "PT5M",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": "5",
                    "TnS": None,
                },
            ),
            (
                "PT6S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": None,
                    "TnS": "6",
                },
            ),
            (
                "PT1H2M",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "1",
                    "TnM": "2",
                    "TnS": None,
                },
            ),
            (
                "PT3H4S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": "3",
                    "TnM": None,
                    "TnS": "4",
                },
            ),
            (
                "PT5M6S",
                {
                    "PnY": None,
                    "PnM": None,
                    "PnW": None,
                    "PnD": None,
                    "TnH": None,
                    "TnM": "5",
                    "TnS": "6",
                },
            ),
        )

        for testtuple in testtuples:
            result = _parse_duration_prescribed_time(testtuple[0])

            self.assertEqual(result, testtuple[1])

    def test_parse_duration_prescribed_time_timeindate(self):
        # Don't allow time components in date half
        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_time("P1Y2M3D4HT54M6S")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_time("P1Y2M3D6ST4H54M")

    def test_parse_duration_prescribed_time_dateintime(self):
        # Don't allow date components in time half
        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_time("P2M3DT1Y4H54M6S")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_time("P1Y2MT3D4H54M6S")

    def test_parse_duration_prescribed_time_outoforder(self):
        # Ensure durations are required to be in the correct order
        # https://bitbucket.org/nielsenb/aniso8601/issues/7/durations-with-time-components-before-t
        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_time("1Y2M3D1SPT1M")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_time("P1Y2M3D2MT1S")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_time("P2M3D1ST1Y1M")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_time("P1Y2M2MT3D1S")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_time("PT1S1H")

    def test_parse_duration_prescribed_time_badstr(self):
        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_time("P1Y")

        with self.assertRaises(ISOFormatError):
            _parse_duration_prescribed_time("P1Y1M")

    def test_parse_duration_combined(self):
        testtuples = (
            (
                "P0003-06-04T12:30:05",
                {
                    "PnY": "0003",
                    "PnM": "06",
                    "PnD": "04",
                    "TnH": "12",
                    "TnM": "30",
                    "TnS": "05",
                },
            ),
            (
                "P0003-06-04T12:30:05,5",
                {
                    "PnY": "0003",
                    "PnM": "06",
                    "PnD": "04",
                    "TnH": "12",
                    "TnM": "30",
                    "TnS": "05.5",
                },
            ),
            (
                "P0003-06-04T12:30:05.5",
                {
                    "PnY": "0003",
                    "PnM": "06",
                    "PnD": "04",
                    "TnH": "12",
                    "TnM": "30",
                    "TnS": "05.5",
                },
            ),
            (
                "P0001-02-03T14:43:59.9999997",
                {
                    "PnY": "0001",
                    "PnM": "02",
                    "PnD": "03",
                    "TnH": "14",
                    "TnM": "43",
                    "TnS": "59.9999997",
                },
            ),
        )

        for testtuple in testtuples:
            result = _parse_duration_combined(testtuple[0])

            self.assertEqual(result, testtuple[1])

    def test_parse_duration_combined_suffixgarbage(self):
        # Don't allow garbage after the duration
        # https://bitbucket.org/nielsenb/aniso8601/issues/9/durations-with-trailing-garbage-are-parsed
        with self.assertRaises(ISOFormatError):
            _parse_duration_combined("P0003-06-04T12:30:05.5asdfasdf")

    def test_has_any_component(self):
        self.assertTrue(_has_any_component("P1Y", ["Y", "M"]))
        self.assertFalse(_has_any_component("P1Y", ["M", "D"]))



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import unittest

import aniso8601
from aniso8601.builders import DatetimeTuple, DateTuple, TimeTuple, TimezoneTuple
from aniso8601.exceptions import ISOFormatError
from aniso8601.resolution import TimeResolution
from aniso8601.tests.compat import mock
from aniso8601.time import (
    _get_time_resolution,
    get_datetime_resolution,
    get_time_resolution,
    parse_datetime,
    parse_time,
)


class TestTimeResolutionFunctions(unittest.TestCase):
    def test_get_time_resolution(self):
        self.assertEqual(get_time_resolution("01:23:45"), TimeResolution.Seconds)
        self.assertEqual(get_time_resolution("24:00:00"), TimeResolution.Seconds)
        self.assertEqual(get_time_resolution("23:21:28,512400"), TimeResolution.Seconds)
        self.assertEqual(get_time_resolution("23:21:28.512400"), TimeResolution.Seconds)
        self.assertEqual(get_time_resolution("01:23"), TimeResolution.Minutes)
        self.assertEqual(get_time_resolution("24:00"), TimeResolution.Minutes)
        self.assertEqual(get_time_resolution("01:23,4567"), TimeResolution.Minutes)
        self.assertEqual(get_time_resolution("01:23.4567"), TimeResolution.Minutes)
        self.assertEqual(get_time_resolution("012345"), TimeResolution.Seconds)
        self.assertEqual(get_time_resolution("240000"), TimeResolution.Seconds)
        self.assertEqual(get_time_resolution("0123"), TimeResolution.Minutes)
        self.assertEqual(get_time_resolution("2400"), TimeResolution.Minutes)
        self.assertEqual(get_time_resolution("01"), TimeResolution.Hours)
        self.assertEqual(get_time_resolution("24"), TimeResolution.Hours)
        self.assertEqual(get_time_resolution("12,5"), TimeResolution.Hours)
        self.assertEqual(get_time_resolution("12.5"), TimeResolution.Hours)
        self.assertEqual(
            get_time_resolution("232128.512400+00:00"), TimeResolution.Seconds
        )
        self.assertEqual(get_time_resolution("0123.4567+00:00"), TimeResolution.Minutes)
        self.assertEqual(get_time_resolution("01.4567+00:00"), TimeResolution.Hours)
        self.assertEqual(get_time_resolution("01:23:45+00:00"), TimeResolution.Seconds)
        self.assertEqual(get_time_resolution("24:00:00+00:00"), TimeResolution.Seconds)
        self.assertEqual(
            get_time_resolution("23:21:28.512400+00:00"), TimeResolution.Seconds
        )
        self.assertEqual(get_time_resolution("01:23+00:00"), TimeResolution.Minutes)
        self.assertEqual(get_time_resolution("24:00+00:00"), TimeResolution.Minutes)
        self.assertEqual(
            get_time_resolution("01:23.4567+00:00"), TimeResolution.Minutes
        )
        self.assertEqual(
            get_time_resolution("23:21:28.512400+11:15"), TimeResolution.Seconds
        )
        self.assertEqual(
            get_time_resolution("23:21:28.512400-12:34"), TimeResolution.Seconds
        )
        self.assertEqual(
            get_time_resolution("23:21:28.512400Z"), TimeResolution.Seconds
        )
        self.assertEqual(
            get_time_resolution("06:14:00.000123Z"), TimeResolution.Seconds
        )

    def test_get_datetime_resolution(self):
        self.assertEqual(
            get_datetime_resolution("2019-06-05T01:03:11.858714"),
            TimeResolution.Seconds,
        )
        self.assertEqual(
            get_datetime_resolution("2019-06-05T01:03:11"), TimeResolution.Seconds
        )
        self.assertEqual(
            get_datetime_resolution("2019-06-05T01:03"), TimeResolution.Minutes
        )
        self.assertEqual(get_datetime_resolution("2019-06-05T01"), TimeResolution.Hours)

    def test_get_time_resolution_badtype(self):
        testtuples = (None, 1, False, 1.234)

        for testtuple in testtuples:
            with self.assertRaises(ValueError):
                get_time_resolution(testtuple)

    def test_get_time_resolution_badstr(self):
        testtuples = ("A6:14:00.000123Z", "06:14:0B", "bad", "")

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                get_time_resolution(testtuple)

    def test_get_time_resolution_internal(self):
        self.assertEqual(
            _get_time_resolution(TimeTuple(hh="01", mm="02", ss="03", tz=None)),
            TimeResolution.Seconds,
        )
        self.assertEqual(
            _get_time_resolution(TimeTuple(hh="01", mm="02", ss=None, tz=None)),
            TimeResolution.Minutes,
        )
        self.assertEqual(
            _get_time_resolution(TimeTuple(hh="01", mm=None, ss=None, tz=None)),
            TimeResolution.Hours,
        )


class TestTimeParserFunctions(unittest.TestCase):
    def test_parse_time(self):
        testtuples = (
            ("01:23:45", {"hh": "01", "mm": "23", "ss": "45", "tz": None}),
            ("24:00:00", {"hh": "24", "mm": "00", "ss": "00", "tz": None}),
            (
                "23:21:28,512400",
                {"hh": "23", "mm": "21", "ss": "28.512400", "tz": None},
            ),
            (
                "23:21:28.512400",
                {"hh": "23", "mm": "21", "ss": "28.512400", "tz": None},
            ),
            (
                "01:03:11.858714",
                {"hh": "01", "mm": "03", "ss": "11.858714", "tz": None},
            ),
            (
                "14:43:59.9999997",
                {"hh": "14", "mm": "43", "ss": "59.9999997", "tz": None},
            ),
            ("01:23", {"hh": "01", "mm": "23", "ss": None, "tz": None}),
            ("24:00", {"hh": "24", "mm": "00", "ss": None, "tz": None}),
            ("01:23,4567", {"hh": "01", "mm": "23.4567", "ss": None, "tz": None}),
            ("01:23.4567", {"hh": "01", "mm": "23.4567", "ss": None, "tz": None}),
            ("012345", {"hh": "01", "mm": "23", "ss": "45", "tz": None}),
            ("240000", {"hh": "24", "mm": "00", "ss": "00", "tz": None}),
            ("232128,512400", {"hh": "23", "mm": "21", "ss": "28.512400", "tz": None}),
            ("232128.512400", {"hh": "23", "mm": "21", "ss": "28.512400", "tz": None}),
            ("010311.858714", {"hh": "01", "mm": "03", "ss": "11.858714", "tz": None}),
            (
                "144359.9999997",
                {"hh": "14", "mm": "43", "ss": "59.9999997", "tz": None},
            ),
            ("0123", {"hh": "01", "mm": "23", "ss": None, "tz": None}),
            ("2400", {"hh": "24", "mm": "00", "ss": None, "tz": None}),
            ("01", {"hh": "01", "mm": None, "ss": None, "tz": None}),
            ("24", {"hh": "24", "mm": None, "ss": None, "tz": None}),
            ("12,5", {"hh": "12.5", "mm": None, "ss": None, "tz": None}),
            ("12.5", {"hh": "12.5", "mm": None, "ss": None, "tz": None}),
            (
                "232128,512400+00:00",
                {
                    "hh": "23",
                    "mm": "21",
                    "ss": "28.512400",
                    "tz": TimezoneTuple(False, None, "00", "00", "+00:00"),
                },
            ),
            (
                "232128.512400+00:00",
                {
                    "hh": "23",
                    "mm": "21",
                    "ss": "28.512400",
                    "tz": TimezoneTuple(False, None, "00", "00", "+00:00"),
                },
            ),
            (
                "0123,4567+00:00",
                {
                    "hh": "01",
                    "mm": "23.4567",
                    "ss": None,
                    "tz": TimezoneTuple(False, None, "00", "00", "+00:00"),
                },
            ),
            (
                "0123.4567+00:00",
                {
                    "hh": "01",
                    "mm": "23.4567",
                    "ss": None,
                    "tz": TimezoneTuple(False, None, "00", "00", "+00:00"),
                },
            ),
            (
                "01,4567+00:00",
                {
                    "hh": "01.4567",
                    "mm": None,
                    "ss": None,
                    "tz": TimezoneTuple(False, None, "00", "00", "+00:00"),
                },
            ),
            (
                "01.4567+00:00",
                {
                    "hh": "01.4567",
                    "mm": None,
                    "ss": None,
                    "tz": TimezoneTuple(False, None, "00", "00", "+00:00"),
                },
            ),
            (
                "01:23:45+00:00",
                {
                    "hh": "01",
                    "mm": "23",
                    "ss": "45",
                    "tz": TimezoneTuple(False, None, "00", "00", "+00:00"),
                },
            ),
            (
                "24:00:00+00:00",
                {
                    "hh": "24",
                    "mm": "00",
                    "ss": "00",
                    "tz": TimezoneTuple(False, None, "00", "00", "+00:00"),
                },
            ),
            (
                "23:21:28.512400+00:00",
                {
                    "hh": "23",
                    "mm": "21",
                    "ss": "28.512400",
                    "tz": TimezoneTuple(False, None, "00", "00", "+00:00"),
                },
            ),
            (
                "01:23+00:00",
                {
                    "hh": "01",
                    "mm": "23",
                    "ss": None,
                    "tz": TimezoneTuple(False, None, "00", "00", "+00:00"),
                },
            ),
            (
                "24:00+00:00",
                {
                    "hh": "24",
                    "mm": "00",
                    "ss": None,
                    "tz": TimezoneTuple(False, None, "00", "00", "+00:00"),
                },
            ),
            (
                "01:23.4567+00:00",
                {
                    "hh": "01",
                    "mm": "23.4567",
                    "ss": None,
                    "tz": TimezoneTuple(False, None, "00", "00", "+00:00"),
                },
            ),
            (
                "23:21:28.512400+11:15",
                {
                    "hh": "23",
                    "mm": "21",
                    "ss": "28.512400",
                    "tz": TimezoneTuple(False, None, "11", "15", "+11:15"),
                },
            ),
            (
                "23:21:28.512400-12:34",
                {
                    "hh": "23",
                    "mm": "21",
                    "ss": "28.512400",
                    "tz": TimezoneTuple(True, None, "12", "34", "-12:34"),
                },
            ),
            (
                "23:21:28.512400Z",
                {
                    "hh": "23",
                    "mm": "21",
                    "ss": "28.512400",
                    "tz": TimezoneTuple(False, True, None, None, "Z"),
                },
            ),
            (
                "06:14:00.000123Z",
                {
                    "hh": "06",
                    "mm": "14",
                    "ss": "00.000123",
                    "tz": TimezoneTuple(False, True, None, None, "Z"),
                },
            ),
        )

        for testtuple in testtuples:
            with mock.patch.object(
                aniso8601.time.PythonTimeBuilder, "build_time"
            ) as mockBuildTime:

                mockBuildTime.return_value = testtuple[1]

                result = parse_time(testtuple[0])

                self.assertEqual(result, testtuple[1])
                mockBuildTime.assert_called_once_with(**testtuple[1])

    def test_parse_time_badtype(self):
        testtuples = (None, 1, False, 1.234)

        for testtuple in testtuples:
            with self.assertRaises(ValueError):
                parse_time(testtuple, builder=None)

    def test_parse_time_badstr(self):
        testtuples = (
            "A6:14:00.000123Z",
            "06:14:0B",
            "06:1 :02",
            "0000,70:24,9",
            "00.27:5332",
            "bad",
            "",
        )

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                parse_time(testtuple, builder=None)

    def test_parse_time_mockbuilder(self):
        mockBuilder = mock.Mock()

        expectedargs = {"hh": "01", "mm": "23", "ss": "45", "tz": None}

        mockBuilder.build_time.return_value = expectedargs

        result = parse_time("01:23:45", builder=mockBuilder)

        self.assertEqual(result, expectedargs)
        mockBuilder.build_time.assert_called_once_with(**expectedargs)

        mockBuilder = mock.Mock()

        expectedargs = {
            "hh": "23",
            "mm": "21",
            "ss": "28.512400",
            "tz": TimezoneTuple(False, None, "00", "00", "+00:00"),
        }

        mockBuilder.build_time.return_value = expectedargs

        result = parse_time("232128.512400+00:00", builder=mockBuilder)

        self.assertEqual(result, expectedargs)
        mockBuilder.build_time.assert_called_once_with(**expectedargs)

        mockBuilder = mock.Mock()

        expectedargs = {
            "hh": "23",
            "mm": "21",
            "ss": "28.512400",
            "tz": TimezoneTuple(False, None, "11", "15", "+11:15"),
        }

        mockBuilder.build_time.return_value = expectedargs

        result = parse_time("23:21:28.512400+11:15", builder=mockBuilder)

        self.assertEqual(result, expectedargs)
        mockBuilder.build_time.assert_called_once_with(**expectedargs)

    def test_parse_datetime(self):
        testtuples = (
            (
                "2019-06-05T01:03:11,858714",
                (
                    DateTuple("2019", "06", "05", None, None, None),
                    TimeTuple("01", "03", "11.858714", None),
                ),
            ),
            (
                "2019-06-05T01:03:11.858714",
                (
                    DateTuple("2019", "06", "05", None, None, None),
                    TimeTuple("01", "03", "11.858714", None),
                ),
            ),
            (
                "1981-04-05T23:21:28.512400Z",
                (
                    DateTuple("1981", "04", "05", None, None, None),
                    TimeTuple(
                        "23",
                        "21",
                        "28.512400",
                        TimezoneTuple(False, True, None, None, "Z"),
                    ),
                ),
            ),
            (
                "1981095T23:21:28.512400-12:34",
                (
                    DateTuple("1981", None, None, None, None, "095"),
                    TimeTuple(
                        "23",
                        "21",
                        "28.512400",
                        TimezoneTuple(True, None, "12", "34", "-12:34"),
                    ),
                ),
            ),
            (
                "19810405T23:21:28+00",
                (
                    DateTuple("1981", "04", "05", None, None, None),
                    TimeTuple(
                        "23", "21", "28", TimezoneTuple(False, None, "00", None, "+00")
                    ),
                ),
            ),
            (
                "19810405T23:21:28+00:00",
                (
                    DateTuple("1981", "04", "05", None, None, None),
                    TimeTuple(
                        "23",
                        "21",
                        "28",
                        TimezoneTuple(False, None, "00", "00", "+00:00"),
                    ),
                ),
            ),
        )

        for testtuple in testtuples:
            with mock.patch.object(
                aniso8601.time.PythonTimeBuilder, "build_datetime"
            ) as mockBuildDateTime:

                mockBuildDateTime.return_value = testtuple[1]

                result = parse_datetime(testtuple[0])

            self.assertEqual(result, testtuple[1])
            mockBuildDateTime.assert_called_once_with(*testtuple[1])

    def test_parse_datetime_spacedelimited(self):
        expectedargs = (
            DateTuple("2004", None, None, "53", "6", None),
            TimeTuple(
                "23", "21", "28.512400", TimezoneTuple(True, None, "12", "34", "-12:34")
            ),
        )

        with mock.patch.object(
            aniso8601.time.PythonTimeBuilder, "build_datetime"
        ) as mockBuildDateTime:

            mockBuildDateTime.return_value = expectedargs

            result = parse_datetime("2004-W53-6 23:21:28.512400-12:34", delimiter=" ")

        self.assertEqual(result, expectedargs)
        mockBuildDateTime.assert_called_once_with(*expectedargs)

    def test_parse_datetime_commadelimited(self):
        expectedargs = (
            DateTuple("1981", "04", "05", None, None, None),
            TimeTuple(
                "23", "21", "28.512400", TimezoneTuple(False, True, None, None, "Z")
            ),
        )

        with mock.patch.object(
            aniso8601.time.PythonTimeBuilder, "build_datetime"
        ) as mockBuildDateTime:

            mockBuildDateTime.return_value = expectedargs

            result = parse_datetime("1981-04-05,23:21:28,512400Z", delimiter=",")

        self.assertEqual(result, expectedargs)
        mockBuildDateTime.assert_called_once_with(*expectedargs)

    def test_parse_datetime_baddelimiter(self):
        testtuples = (
            "1981-04-05,23:21:28,512400Z",
            "2004-W53-6 23:21:28.512400-12:3",
            "1981040523:21:28",
        )

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                parse_datetime(testtuple, builder=None)

    def test_parse_datetime_badtype(self):
        testtuples = (None, 1, False, 1.234)

        for testtuple in testtuples:
            with self.assertRaises(ValueError):
                parse_datetime(testtuple, builder=None)

    def test_parse_datetime_badstr(self):
        testtuples = (
            "1981-04-05TA6:14:00.000123Z",
            "2004-W53-6T06:14:0B",
            "2014-01-230T23:21:28+00",
            "201401230T01:03:11.858714",
            "9999 W53T49",
            "9T0000,70:24,9",
            "bad",
            "",
        )

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                parse_datetime(testtuple, builder=None)

    def test_parse_datetime_mockbuilder(self):
        mockBuilder = mock.Mock()

        expectedargs = (
            DateTuple("1981", None, None, None, None, "095"),
            TimeTuple(
                "23", "21", "28.512400", TimezoneTuple(True, None, "12", "34", "-12:34")
            ),
        )

        mockBuilder.build_datetime.return_value = expectedargs

        result = parse_datetime("1981095T23:21:28.512400-12:34", builder=mockBuilder)

        self.assertEqual(result, expectedargs)
        mockBuilder.build_datetime.assert_called_once_with(*expectedargs)



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import unittest

from aniso8601.decimalfraction import normalize


class TestDecimalFractionFunctions(unittest.TestCase):
    def test_normalize(self):
        self.assertEqual(normalize(""), "")
        self.assertEqual(normalize("12.34"), "12.34")
        self.assertEqual(normalize("123,45"), "123.45")
        self.assertEqual(normalize("123,45,67"), "123.45.67")



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from aniso8601.date import get_date_resolution, parse_date
from aniso8601.duration import get_duration_resolution, parse_duration
from aniso8601.interval import (
    get_interval_resolution,
    get_repeating_interval_resolution,
    parse_interval,
    parse_repeating_interval,
)

# Import the main parsing functions so they are readily available
from aniso8601.time import (
    get_datetime_resolution,
    get_time_resolution,
    parse_datetime,
    parse_time,
)

__version__ = "10.0.0"



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import calendar
from collections import namedtuple

from aniso8601.exceptions import (
    DayOutOfBoundsError,
    HoursOutOfBoundsError,
    ISOFormatError,
    LeapSecondError,
    MidnightBoundsError,
    MinutesOutOfBoundsError,
    MonthOutOfBoundsError,
    SecondsOutOfBoundsError,
    WeekOutOfBoundsError,
    YearOutOfBoundsError,
)

DateTuple = namedtuple("Date", ["YYYY", "MM", "DD", "Www", "D", "DDD"])
TimeTuple = namedtuple("Time", ["hh", "mm", "ss", "tz"])
DatetimeTuple = namedtuple("Datetime", ["date", "time"])
DurationTuple = namedtuple(
    "Duration", ["PnY", "PnM", "PnW", "PnD", "TnH", "TnM", "TnS"]
)
IntervalTuple = namedtuple("Interval", ["start", "end", "duration"])
RepeatingIntervalTuple = namedtuple("RepeatingInterval", ["R", "Rnn", "interval"])
TimezoneTuple = namedtuple("Timezone", ["negative", "Z", "hh", "mm", "name"])

Limit = namedtuple(
    "Limit",
    [
        "casterrorstring",
        "min",
        "max",
        "rangeexception",
        "rangeerrorstring",
        "rangefunc",
    ],
)


def cast(
    value,
    castfunction,
    caughtexceptions=(ValueError,),
    thrownexception=ISOFormatError,
    thrownmessage=None,
):
    try:
        result = castfunction(value)
    except caughtexceptions:
        raise thrownexception(thrownmessage)

    return result


def range_check(valuestr, limit):
    # Returns cast value if in range, raises defined exceptions on failure
    if valuestr is None:
        return None

    if "." in valuestr:
        castfunc = float
    else:
        castfunc = int

    value = cast(valuestr, castfunc, thrownmessage=limit.casterrorstring)

    if limit.min is not None and value < limit.min:
        raise limit.rangeexception(limit.rangeerrorstring)

    if limit.max is not None and value > limit.max:
        raise limit.rangeexception(limit.rangeerrorstring)

    return value


class BaseTimeBuilder(object):
    # Limit tuple format cast function, cast error string,
    # lower limit, upper limit, limit error string
    DATE_YYYY_LIMIT = Limit(
        "Invalid year string.",
        0000,
        9999,
        YearOutOfBoundsError,
        "Year must be between 1..9999.",
        range_check,
    )
    DATE_MM_LIMIT = Limit(
        "Invalid month string.",
        1,
        12,
        MonthOutOfBoundsError,
        "Month must be between 1..12.",
        range_check,
    )
    DATE_DD_LIMIT = Limit(
        "Invalid day string.",
        1,
        31,
        DayOutOfBoundsError,
        "Day must be between 1..31.",
        range_check,
    )
    DATE_WWW_LIMIT = Limit(
        "Invalid week string.",
        1,
        53,
        WeekOutOfBoundsError,
        "Week number must be between 1..53.",
        range_check,
    )
    DATE_D_LIMIT = Limit(
        "Invalid weekday string.",
        1,
        7,
        DayOutOfBoundsError,
        "Weekday number must be between 1..7.",
        range_check,
    )
    DATE_DDD_LIMIT = Limit(
        "Invalid ordinal day string.",
        1,
        366,
        DayOutOfBoundsError,
        "Ordinal day must be between 1..366.",
        range_check,
    )
    TIME_HH_LIMIT = Limit(
        "Invalid hour string.",
        0,
        24,
        HoursOutOfBoundsError,
        "Hour must be between 0..24 with 24 representing midnight.",
        range_check,
    )
    TIME_MM_LIMIT = Limit(
        "Invalid minute string.",
        0,
        59,
        MinutesOutOfBoundsError,
        "Minute must be between 0..59.",
        range_check,
    )
    TIME_SS_LIMIT = Limit(
        "Invalid second string.",
        0,
        60,
        SecondsOutOfBoundsError,
        "Second must be between 0..60 with 60 representing a leap second.",
        range_check,
    )
    TZ_HH_LIMIT = Limit(
        "Invalid timezone hour string.",
        0,
        23,
        HoursOutOfBoundsError,
        "Hour must be between 0..23.",
        range_check,
    )
    TZ_MM_LIMIT = Limit(
        "Invalid timezone minute string.",
        0,
        59,
        MinutesOutOfBoundsError,
        "Minute must be between 0..59.",
        range_check,
    )
    DURATION_PNY_LIMIT = Limit(
        "Invalid year duration string.",
        0,
        None,
        ISOFormatError,
        "Duration years component must be positive.",
        range_check,
    )
    DURATION_PNM_LIMIT = Limit(
        "Invalid month duration string.",
        0,
        None,
        ISOFormatError,
        "Duration months component must be positive.",
        range_check,
    )
    DURATION_PNW_LIMIT = Limit(
        "Invalid week duration string.",
        0,
        None,
        ISOFormatError,
        "Duration weeks component must be positive.",
        range_check,
    )
    DURATION_PND_LIMIT = Limit(
        "Invalid day duration string.",
        0,
        None,
        ISOFormatError,
        "Duration days component must be positive.",
        range_check,
    )
    DURATION_TNH_LIMIT = Limit(
        "Invalid hour duration string.",
        0,
        None,
        ISOFormatError,
        "Duration hours component must be positive.",
        range_check,
    )
    DURATION_TNM_LIMIT = Limit(
        "Invalid minute duration string.",
        0,
        None,
        ISOFormatError,
        "Duration minutes component must be positive.",
        range_check,
    )
    DURATION_TNS_LIMIT = Limit(
        "Invalid second duration string.",
        0,
        None,
        ISOFormatError,
        "Duration seconds component must be positive.",
        range_check,
    )
    INTERVAL_RNN_LIMIT = Limit(
        "Invalid duration repetition string.",
        0,
        None,
        ISOFormatError,
        "Duration repetition count must be positive.",
        range_check,
    )

    DATE_RANGE_DICT = {
        "YYYY": DATE_YYYY_LIMIT,
        "MM": DATE_MM_LIMIT,
        "DD": DATE_DD_LIMIT,
        "Www": DATE_WWW_LIMIT,
        "D": DATE_D_LIMIT,
        "DDD": DATE_DDD_LIMIT,
    }

    TIME_RANGE_DICT = {"hh": TIME_HH_LIMIT, "mm": TIME_MM_LIMIT, "ss": TIME_SS_LIMIT}

    DURATION_RANGE_DICT = {
        "PnY": DURATION_PNY_LIMIT,
        "PnM": DURATION_PNM_LIMIT,
        "PnW": DURATION_PNW_LIMIT,
        "PnD": DURATION_PND_LIMIT,
        "TnH": DURATION_TNH_LIMIT,
        "TnM": DURATION_TNM_LIMIT,
        "TnS": DURATION_TNS_LIMIT,
    }

    REPEATING_INTERVAL_RANGE_DICT = {"Rnn": INTERVAL_RNN_LIMIT}

    TIMEZONE_RANGE_DICT = {"hh": TZ_HH_LIMIT, "mm": TZ_MM_LIMIT}

    LEAP_SECONDS_SUPPORTED = False

    @classmethod
    def build_date(cls, YYYY=None, MM=None, DD=None, Www=None, D=None, DDD=None):
        raise NotImplementedError

    @classmethod
    def build_time(cls, hh=None, mm=None, ss=None, tz=None):
        raise NotImplementedError

    @classmethod
    def build_datetime(cls, date, time):
        raise NotImplementedError

    @classmethod
    def build_duration(
        cls, PnY=None, PnM=None, PnW=None, PnD=None, TnH=None, TnM=None, TnS=None
    ):
        raise NotImplementedError

    @classmethod
    def build_interval(cls, start=None, end=None, duration=None):
        # start, end, and duration are all tuples
        raise NotImplementedError

    @classmethod
    def build_repeating_interval(cls, R=None, Rnn=None, interval=None):
        # interval is a tuple
        raise NotImplementedError

    @classmethod
    def build_timezone(cls, negative=None, Z=None, hh=None, mm=None, name=""):
        raise NotImplementedError

    @classmethod
    def range_check_date(
        cls, YYYY=None, MM=None, DD=None, Www=None, D=None, DDD=None, rangedict=None
    ):
        if rangedict is None:
            rangedict = cls.DATE_RANGE_DICT

        if "YYYY" in rangedict:
            YYYY = rangedict["YYYY"].rangefunc(YYYY, rangedict["YYYY"])

        if "MM" in rangedict:
            MM = rangedict["MM"].rangefunc(MM, rangedict["MM"])

        if "DD" in rangedict:
            DD = rangedict["DD"].rangefunc(DD, rangedict["DD"])

        if "Www" in rangedict:
            Www = rangedict["Www"].rangefunc(Www, rangedict["Www"])

        if "D" in rangedict:
            D = rangedict["D"].rangefunc(D, rangedict["D"])

        if "DDD" in rangedict:
            DDD = rangedict["DDD"].rangefunc(DDD, rangedict["DDD"])

        if DD is not None:
            # Check calendar
            if DD > calendar.monthrange(YYYY, MM)[1]:
                raise DayOutOfBoundsError(
                    "{0} is out of range for {1}-{2}".format(DD, YYYY, MM)
                )

        if DDD is not None:
            if calendar.isleap(YYYY) is False and DDD == 366:
                raise DayOutOfBoundsError(
                    "{0} is only valid for leap year.".format(DDD)
                )

        return (YYYY, MM, DD, Www, D, DDD)

    @classmethod
    def range_check_time(cls, hh=None, mm=None, ss=None, tz=None, rangedict=None):
        # Used for midnight and leap second handling
        midnight = False  # Handle hh = '24' specially

        if rangedict is None:
            rangedict = cls.TIME_RANGE_DICT

        if "hh" in rangedict:
            try:
                hh = rangedict["hh"].rangefunc(hh, rangedict["hh"])
            except HoursOutOfBoundsError as e:
                if float(hh) > 24 and float(hh) < 25:
                    raise MidnightBoundsError("Hour 24 may only represent midnight.")

                raise e

        if "mm" in rangedict:
            mm = rangedict["mm"].rangefunc(mm, rangedict["mm"])

        if "ss" in rangedict:
            ss = rangedict["ss"].rangefunc(ss, rangedict["ss"])

        if hh is not None and hh == 24:
            midnight = True

        # Handle midnight range
        if midnight is True and (
            (mm is not None and mm != 0) or (ss is not None and ss != 0)
        ):
            raise MidnightBoundsError("Hour 24 may only represent midnight.")

        if cls.LEAP_SECONDS_SUPPORTED is True:
            if hh != 23 and mm != 59 and ss == 60:
                raise cls.TIME_SS_LIMIT.rangeexception(
                    cls.TIME_SS_LIMIT.rangeerrorstring
                )
        else:
            if hh == 23 and mm == 59 and ss == 60:
                # https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
                raise LeapSecondError("Leap seconds are not supported.")

            if ss == 60:
                raise cls.TIME_SS_LIMIT.rangeexception(
                    cls.TIME_SS_LIMIT.rangeerrorstring
                )

        return (hh, mm, ss, tz)

    @classmethod
    def range_check_duration(
        cls,
        PnY=None,
        PnM=None,
        PnW=None,
        PnD=None,
        TnH=None,
        TnM=None,
        TnS=None,
        rangedict=None,
    ):
        if rangedict is None:
            rangedict = cls.DURATION_RANGE_DICT

        if "PnY" in rangedict:
            PnY = rangedict["PnY"].rangefunc(PnY, rangedict["PnY"])

        if "PnM" in rangedict:
            PnM = rangedict["PnM"].rangefunc(PnM, rangedict["PnM"])

        if "PnW" in rangedict:
            PnW = rangedict["PnW"].rangefunc(PnW, rangedict["PnW"])

        if "PnD" in rangedict:
            PnD = rangedict["PnD"].rangefunc(PnD, rangedict["PnD"])

        if "TnH" in rangedict:
            TnH = rangedict["TnH"].rangefunc(TnH, rangedict["TnH"])

        if "TnM" in rangedict:
            TnM = rangedict["TnM"].rangefunc(TnM, rangedict["TnM"])

        if "TnS" in rangedict:
            TnS = rangedict["TnS"].rangefunc(TnS, rangedict["TnS"])

        return (PnY, PnM, PnW, PnD, TnH, TnM, TnS)

    @classmethod
    def range_check_repeating_interval(
        cls, R=None, Rnn=None, interval=None, rangedict=None
    ):
        if rangedict is None:
            rangedict = cls.REPEATING_INTERVAL_RANGE_DICT

        if "Rnn" in rangedict:
            Rnn = rangedict["Rnn"].rangefunc(Rnn, rangedict["Rnn"])

        return (R, Rnn, interval)

    @classmethod
    def range_check_timezone(
        cls, negative=None, Z=None, hh=None, mm=None, name="", rangedict=None
    ):
        if rangedict is None:
            rangedict = cls.TIMEZONE_RANGE_DICT

        if "hh" in rangedict:
            hh = rangedict["hh"].rangefunc(hh, rangedict["hh"])

        if "mm" in rangedict:
            mm = rangedict["mm"].rangefunc(mm, rangedict["mm"])

        return (negative, Z, hh, mm, name)

    @classmethod
    def _build_object(cls, parsetuple):
        # Given a TupleBuilder tuple, build the correct object
        if isinstance(parsetuple, DateTuple):
            return cls.build_date(
                YYYY=parsetuple.YYYY,
                MM=parsetuple.MM,
                DD=parsetuple.DD,
                Www=parsetuple.Www,
                D=parsetuple.D,
                DDD=parsetuple.DDD,
            )

        if isinstance(parsetuple, TimeTuple):
            return cls.build_time(
                hh=parsetuple.hh, mm=parsetuple.mm, ss=parsetuple.ss, tz=parsetuple.tz
            )

        if isinstance(parsetuple, DatetimeTuple):
            return cls.build_datetime(parsetuple.date, parsetuple.time)

        if isinstance(parsetuple, DurationTuple):
            return cls.build_duration(
                PnY=parsetuple.PnY,
                PnM=parsetuple.PnM,
                PnW=parsetuple.PnW,
                PnD=parsetuple.PnD,
                TnH=parsetuple.TnH,
                TnM=parsetuple.TnM,
                TnS=parsetuple.TnS,
            )

        if isinstance(parsetuple, IntervalTuple):
            return cls.build_interval(
                start=parsetuple.start, end=parsetuple.end, duration=parsetuple.duration
            )

        if isinstance(parsetuple, RepeatingIntervalTuple):
            return cls.build_repeating_interval(
                R=parsetuple.R, Rnn=parsetuple.Rnn, interval=parsetuple.interval
            )

        return cls.build_timezone(
            negative=parsetuple.negative,
            Z=parsetuple.Z,
            hh=parsetuple.hh,
            mm=parsetuple.mm,
            name=parsetuple.name,
        )

    @classmethod
    def _is_interval_end_concise(cls, endtuple):
        if isinstance(endtuple, TimeTuple):
            return True

        if isinstance(endtuple, DatetimeTuple):
            enddatetuple = endtuple.date
        else:
            enddatetuple = endtuple

        if enddatetuple.YYYY is None:
            return True

        return False

    @classmethod
    def _combine_concise_interval_tuples(cls, starttuple, conciseendtuple):
        starttimetuple = None
        startdatetuple = None

        endtimetuple = None
        enddatetuple = None

        if isinstance(starttuple, DateTuple):
            startdatetuple = starttuple
        else:
            # Start is a datetime
            starttimetuple = starttuple.time
            startdatetuple = starttuple.date

        if isinstance(conciseendtuple, DateTuple):
            enddatetuple = conciseendtuple
        elif isinstance(conciseendtuple, DatetimeTuple):
            enddatetuple = conciseendtuple.date
            endtimetuple = conciseendtuple.time
        else:
            # Time
            endtimetuple = conciseendtuple

        if enddatetuple is not None:
            if enddatetuple.YYYY is None and enddatetuple.MM is None:
                newenddatetuple = DateTuple(
                    YYYY=startdatetuple.YYYY,
                    MM=startdatetuple.MM,
                    DD=enddatetuple.DD,
                    Www=enddatetuple.Www,
                    D=enddatetuple.D,
                    DDD=enddatetuple.DDD,
                )
            else:
                newenddatetuple = DateTuple(
                    YYYY=startdatetuple.YYYY,
                    MM=enddatetuple.MM,
                    DD=enddatetuple.DD,
                    Www=enddatetuple.Www,
                    D=enddatetuple.D,
                    DDD=enddatetuple.DDD,
                )

            if endtimetuple is None:
                return newenddatetuple

        if (starttimetuple is not None and starttimetuple.tz is not None) and (
            endtimetuple is not None and endtimetuple.tz != starttimetuple.tz
        ):
            # Copy the timezone across
            endtimetuple = TimeTuple(
                hh=endtimetuple.hh,
                mm=endtimetuple.mm,
                ss=endtimetuple.ss,
                tz=starttimetuple.tz,
            )

        if enddatetuple is not None and endtimetuple is not None:
            return TupleBuilder.build_datetime(newenddatetuple, endtimetuple)

        return TupleBuilder.build_datetime(startdatetuple, endtimetuple)


class TupleBuilder(BaseTimeBuilder):
    # Builder used to return the arguments as a tuple, cleans up some parse methods
    @classmethod
    def build_date(cls, YYYY=None, MM=None, DD=None, Www=None, D=None, DDD=None):

        return DateTuple(YYYY, MM, DD, Www, D, DDD)

    @classmethod
    def build_time(cls, hh=None, mm=None, ss=None, tz=None):
        return TimeTuple(hh, mm, ss, tz)

    @classmethod
    def build_datetime(cls, date, time):
        return DatetimeTuple(date, time)

    @classmethod
    def build_duration(
        cls, PnY=None, PnM=None, PnW=None, PnD=None, TnH=None, TnM=None, TnS=None
    ):

        return DurationTuple(PnY, PnM, PnW, PnD, TnH, TnM, TnS)

    @classmethod
    def build_interval(cls, start=None, end=None, duration=None):
        return IntervalTuple(start, end, duration)

    @classmethod
    def build_repeating_interval(cls, R=None, Rnn=None, interval=None):
        return RepeatingIntervalTuple(R, Rnn, interval)

    @classmethod
    def build_timezone(cls, negative=None, Z=None, hh=None, mm=None, name=""):
        return TimezoneTuple(negative, Z, hh, mm, name)



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import unittest

import aniso8601
from aniso8601.exceptions import ISOFormatError
from aniso8601.tests.compat import mock
from aniso8601.timezone import parse_timezone


class TestTimezoneParserFunctions(unittest.TestCase):
    def test_parse_timezone(self):
        testtuples = (
            ("Z", {"negative": False, "Z": True, "name": "Z"}),
            ("+00:00", {"negative": False, "hh": "00", "mm": "00", "name": "+00:00"}),
            ("+01:00", {"negative": False, "hh": "01", "mm": "00", "name": "+01:00"}),
            ("-01:00", {"negative": True, "hh": "01", "mm": "00", "name": "-01:00"}),
            ("+00:12", {"negative": False, "hh": "00", "mm": "12", "name": "+00:12"}),
            ("+01:23", {"negative": False, "hh": "01", "mm": "23", "name": "+01:23"}),
            ("-01:23", {"negative": True, "hh": "01", "mm": "23", "name": "-01:23"}),
            ("+0000", {"negative": False, "hh": "00", "mm": "00", "name": "+0000"}),
            ("+0100", {"negative": False, "hh": "01", "mm": "00", "name": "+0100"}),
            ("-0100", {"negative": True, "hh": "01", "mm": "00", "name": "-0100"}),
            ("+0012", {"negative": False, "hh": "00", "mm": "12", "name": "+0012"}),
            ("+0123", {"negative": False, "hh": "01", "mm": "23", "name": "+0123"}),
            ("-0123", {"negative": True, "hh": "01", "mm": "23", "name": "-0123"}),
            ("+00", {"negative": False, "hh": "00", "mm": None, "name": "+00"}),
            ("+01", {"negative": False, "hh": "01", "mm": None, "name": "+01"}),
            ("-01", {"negative": True, "hh": "01", "mm": None, "name": "-01"}),
            ("+12", {"negative": False, "hh": "12", "mm": None, "name": "+12"}),
            ("-12", {"negative": True, "hh": "12", "mm": None, "name": "-12"}),
        )

        for testtuple in testtuples:
            with mock.patch.object(
                aniso8601.timezone.PythonTimeBuilder, "build_timezone"
            ) as mockBuildTimezone:

                mockBuildTimezone.return_value = testtuple[1]

                result = parse_timezone(testtuple[0])

                self.assertEqual(result, testtuple[1])
                mockBuildTimezone.assert_called_once_with(**testtuple[1])

    def test_parse_timezone_badtype(self):
        testtuples = (None, 1, False, 1.234)

        for testtuple in testtuples:
            with self.assertRaises(ValueError):
                parse_timezone(testtuple, builder=None)

    def test_parse_timezone_badstr(self):
        testtuples = (
            "+1",
            "-00",
            "-0000",
            "-00:00",
            "01",
            "0123",
            "@12:34",
            "Y",
            " Z",
            "Z ",
            " Z ",
            "bad",
            "",
        )

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                parse_timezone(testtuple, builder=None)

    def test_parse_timezone_mockbuilder(self):
        mockBuilder = mock.Mock()

        expectedargs = {"negative": False, "Z": True, "name": "Z"}

        mockBuilder.build_timezone.return_value = expectedargs

        result = parse_timezone("Z", builder=mockBuilder)

        self.assertEqual(result, expectedargs)
        mockBuilder.build_timezone.assert_called_once_with(**expectedargs)

        mockBuilder = mock.Mock()

        expectedargs = {"negative": False, "hh": "00", "mm": "00", "name": "+00:00"}

        mockBuilder.build_timezone.return_value = expectedargs

        result = parse_timezone("+00:00", builder=mockBuilder)

        self.assertEqual(result, expectedargs)
        mockBuilder.build_timezone.assert_called_once_with(**expectedargs)

        mockBuilder = mock.Mock()

        expectedargs = {"negative": True, "hh": "01", "mm": "23", "name": "-01:23"}

        mockBuilder.build_timezone.return_value = expectedargs

        result = parse_timezone("-01:23", builder=mockBuilder)

        self.assertEqual(result, expectedargs)
        mockBuilder.build_timezone.assert_called_once_with(**expectedargs)

    def test_parse_timezone_negativezero(self):
        # A 0 offset cannot be negative
        with self.assertRaises(ISOFormatError):
            parse_timezone("-00:00", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_timezone("-0000", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_timezone("-00", builder=None)



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.


class ISOFormatError(ValueError):
    """Raised when ISO 8601 string fails a format check."""


class RangeCheckError(ValueError):
    """Parent type of range check errors."""


class YearOutOfBoundsError(RangeCheckError):
    """Raised when year exceeds limits."""


class MonthOutOfBoundsError(RangeCheckError):
    """Raised when month is outside of 1..12."""


class WeekOutOfBoundsError(RangeCheckError):
    """Raised when week exceeds a year."""


class DayOutOfBoundsError(RangeCheckError):
    """Raised when day is outside of 1..365, 1..366 for leap year."""


class HoursOutOfBoundsError(RangeCheckError):
    """Raise when parsed hours are greater than 24."""


class MinutesOutOfBoundsError(RangeCheckError):
    """Raise when parsed seconds are greater than 60."""


class SecondsOutOfBoundsError(RangeCheckError):
    """Raise when parsed seconds are greater than 60."""


class MidnightBoundsError(RangeCheckError):
    """Raise when parsed time has an hour of 24 but is not midnight."""


class LeapSecondError(RangeCheckError):
    """Raised when attempting to parse a leap second"""
